--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2024 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/liveconfig.lua
-- LiveConfig resources, written in LUA
-- ---------------------------------------------------------------------------

-- This program is run on every start of LiveConfig. It registeres all
-- LC.* modules and detects the operating system / distribution.
-- If this program can't be run, LiveConfig will not be started.

-- Register all modules (order matters!)
LC.distribution = require("distribution")
LC.hooks = require("hooks")
LC.users = require("users")
LC.cron = require("cron")
LC.web = require("web")
LC.web.load("apache")
LC.web.load("nginx")
LC.dns = require("dns")
LC.dns.load("bind")
LC.smtp = require("smtp")
LC.smtp.load("postfix")
LC.popimap = require("popimap")
LC.popimap.load("dovecot")
LC.spamassassin = require("spamassassin")
LC.dkim = require("dkim")
LC.db = require("db")
LC.db.load("mysql")
LC.ftp = require("ftp")
LC.ftp.load("proftpd")
LC.ftp.load("vsftpd")
LC.backup = require("backup")

-- default SSL cipher string
-- current value: v4.0 (intermediate) from Mozilla OpSec (https://wiki.mozilla.org/Security/Server_Side_TLS)
LC.liveconfig.DEFAULT_SSL_CIPHERS = "ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305:DHE-RSA-AES128-GCM-SHA256:DHE-RSA-AES256-GCM-SHA384:DHE-RSA-CHACHA20-POLY1305:ECDHE-ECDSA-AES128-SHA256:ECDHE-RSA-AES128-SHA256:ECDHE-ECDSA-AES128-SHA:ECDHE-RSA-AES128-SHA:ECDHE-ECDSA-AES256-SHA384:ECDHE-RSA-AES256-SHA384:ECDHE-ECDSA-AES256-SHA:ECDHE-RSA-AES256-SHA:DHE-RSA-AES128-SHA256:DHE-RSA-AES256-SHA256:AES128-GCM-SHA256:AES256-GCM-SHA384:AES128-SHA256:AES256-SHA256:AES128-SHA:AES256-SHA:DES-CBC3-SHA";

-- modern/strong SSL ciphers, might break some ancient clients:
LC.liveconfig.DEFAULT_SSL_STRONG_CIPHERS = "ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305:DHE-RSA-AES128-GCM-SHA256:DHE-RSA-AES256-GCM-SHA384"
                                           
-- Define some helper functions

-- ---------------------------------------------------------------------------
-- LC.liveconfig.writeHeader()
-- ---------------------------------------------------------------------------
function LC.liveconfig.writeHeader(fp, comment)
  if comment == nil then comment = '#' end
  fp:write([[
]], comment, [[  _    _          ___           __ _     (R)
]], comment, [[ | |  (_)_ _____ / __|___ _ _  / _(_)__ _
]], comment, [[ | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
]], comment, [[ |____|_|\_/\___|\___\___/_||_|_| |_\__, |
]], comment, [[                                    |___/
]], comment, [[ Copyright (c) 2009-2024 LiveConfig GmbH.
]], comment, [[ ----------------------------------------------------------------------------
]])
end

-- ---------------------------------------------------------------------------
-- LC.liveconfig.writeFooter()
-- ---------------------------------------------------------------------------
function LC.liveconfig.writeFooter(fp, comment)
  if comment == nil then comment = '#' end
  fp:write([[
]], comment, [[ <EOF>-----------------------------------------------------------------------
]])
end

-- ---------------------------------------------------------------------------
-- LC.liveconfig.readStatus()
-- ---------------------------------------------------------------------------
function LC.liveconfig.readStatus(name)
  local fh, errmsg, errno = io.open(name, "r")
  if fh == nil then
    LC.log.print(LC.log.INFO, "Can't open '", name, "' for reading: ", errmsg)
    return nil
  end

  local tab = {}
  local line
  for line in fh:lines() do repeat
    local m = string.match(line, "# LC_CONFIG_ID:%s*([^\n]+)")
    if m then tab.id = m break end
    m = string.match(line, "# LC_CONFIG_REVISION:%s*(%d+)")
    if m then tab.revision = m break end
    m = string.match(line, "# LC_CONFIG_DATE:%s*(%d%d%d%d%-%d%d%-%d%d %d%d:%d%d:%d%d)")
    if m then tab.date = m break end
  until true end
  fh:close()
  return tab
end

-- ---------------------------------------------------------------------------
-- LC.liveconfig.writeStatus()
-- ---------------------------------------------------------------------------
function LC.liveconfig.writeStatus(name, id, rev, dat)
  local fh, errmsg, errno = io.open(name, "w")
  if fh == nil then
    LC.log.print(LC.log.INFO, "Can't open '", name, "' for writing: ", errmsg)
    return false
  end

  -- write header
  LC.liveconfig.writeHeader(fh)

  -- write information
  fh:write([[
# This is a status file for the LiveConfig Server Control Panel.
# DO NOT MAKE ANY CHANGES HERE! This file is only maintained by LiveConfig.
#
]])

  -- write id, revision and date
  fh:write('# LC_CONFIG_ID: ', id, '\n')
  fh:write('# LC_CONFIG_REVISION: ', rev, '\n')
  fh:write('# LC_CONFIG_DATE: ', dat, '\n')
  fh:write('#\n')

  -- write footer
  LC.liveconfig.writeFooter(fh)

  fh:close()
  return true
end

-- ---------------------------------------------------------------------------
-- LC.liveconfig.inTable()
-- ---------------------------------------------------------------------------
function LC.liveconfig.inTable(needle, haystack)
  for _,i in pairs(haystack) do
    if (i==needle) then return true end
  end
  return false
end

-- ---------------------------------------------------------------------------
-- LC.liveconfig.listAppend()
-- Append value to quoted (!!) list of strings
-- If not quoted, discard content
-- ---------------------------------------------------------------------------
function LC.liveconfig.listAppend(list, str)
  if list == nil then
    return "\"" .. str .. "\""
  else
    local lst = {}, tk
    local found = false
    if list ~= "" then list = string.match(list, "^\"(.*)\"$") end
    if list == nil then list = "" end
    for tk in string.gmatch(list, "[^,%s]+") do
      if tk == str then found = true end
      table.insert(lst, tk)
    end
    if not found then table.insert(lst, str) end
    return "\"" .. table.concat(lst, ",") .. "\""
  end
end

-- ---------------------------------------------------------------------------
-- LC.liveconfig.updateSysconfig()
-- Update system configuration in /etc/liveconfig/sysconfig
-- ---------------------------------------------------------------------------
function LC.liveconfig.updateSysconfig()
  -- this function is only called by the LiveConfig process!
  local fh, errmsg
  local oldvalues = { }

  -- read current values
  if LC.fs.is_file(LC.liveconfig.sysconfdir .. "/sysconfig") then
    fh, errmsg = io.open(LC.liveconfig.sysconfdir .. "/sysconfig", "r")
    if fh == nil then
      -- if this file exists, it should be readable!
      LC.log.print(LC.log.INFO, "Can't open '" .. LC.liveconfig.sysconfdir .. "/sysconfig' for reading: ", errmsg)
      return false
    end
    local line
    while fh ~= nil do
      line = fh:read()
      if line == nil then break end
      local k,v = string.match(line, "^([^=]*)=(.*)")
      if k ~= nil and v ~= nil then
        oldvalues[k] = v
      end
    end
    fh:close()
  end

  local cfgversion = oldvalues['LC_CFGVERSION']
  if cfgversion == nil then cfgversion = 1 end

  -- check for upgrade tasks!
  if tonumber(cfgversion) < 2 then
    -- fix permissions for web directories
    if not LC.fs.is_file(LC.liveconfig.sysconfdir .. "/sysconfig") then
      -- create preliminary sysconfig file to make "fix-permissions" script work
      fh, errmsg = io.open(LC.liveconfig.sysconfdir .. "/sysconfig", "w")
      if fh == nil then
        LC.log.print(LC.log.INFO, "Can't open '" .. LC.liveconfig.sysconfdir .. "/sysconfig' for writing: ", errmsg)
        return false
      end
      fh:write("LC_WEBROOT=", LC.web.getWebRoot(), "\n")
      fh:write("LC_CFGVERSION=", cfgversion, "\n")
      fh:close()
    end
    cfgversion = 2
  end

  if tonumber(cfgversion) == 2 then
    -- fix postfix virtual_domains file
    LC.log.print(LC.log.INFO, "Running '", LC.liveconfig.libexecdir, "/lcservice.sh check-postfix-domains'")
    local rc = LC.exec(LC.liveconfig.libexecdir .. '/lcservice.sh check-postfix-domains')
    -- increase cfg version
    cfgversion = 3
  end

  if tonumber(cfgversion) == 3 then
    -- fix permissions for ~/htdocs/cgi-bin/ (see issue #74)
    -- => now done with rev 5
    -- increase cfg version
    cfgversion = 4
  end

  if tonumber(cfgversion) == 4 then
    -- fix AWStats configuration
    LC.log.print(LC.log.INFO, "Running '", LC.liveconfig.libexecdir, "/lcservice.sh fix-awstats'")
    local rc = LC.exec(LC.liveconfig.libexecdir .. '/lcservice.sh fix-awstats')
    -- increase cfg version
    cfgversion = 5
  end

  if tonumber(cfgversion) == 5 then
    -- fix permissions for ~/logs/
    -- create ~/logs/php/
    -- => now done with rev 6
    -- increase cfg version
    cfgversion = 6
  end

  if tonumber(cfgversion) == 6 then
    -- fix permissions for ~/apps/* to work (again) with suPHP
    -- => now done with rev 8
    -- increase cfg version
    cfgversion = 7
  end

  if tonumber(cfgversion) == 7 then
    -- only a change in liveconfig.conf (lccp_socket); done by upgrade script
    -- increase cfg version
    cfgversion = 8
  end

  if tonumber(cfgversion) == 8 then
    -- fix permissions for ~/conf/php5/php.ini (set "immutable" flag)
    LC.log.print(LC.log.INFO, "Running '", LC.liveconfig.libexecdir, "/lcservice.sh fix-permissions'")
    local rc = LC.exec(LC.liveconfig.libexecdir .. '/lcservice.sh fix-permissions')
    -- increase cfg version
    cfgversion = 9
  end

  if tonumber(cfgversion) == 9 then
    -- fix permissions for /etc/vsftpd/users (if existing)
    if LC.fs.is_dir("/etc/vsftpd/users") then
      LC.fs.setperm("/etc/vsftpd/users", "0750", "root", "root")
    end
    cfgversion = 10
  end

  if tonumber(cfgversion) == 10 then
    -- fix CustomLog directive in Apache
    -- => modified again with rev 12
    cfgversion = 11
  end

  if tonumber(cfgversion) == 11 then
    -- create /etc/liveconfig/lclogparse.conf if Postfix is managed by LiveConfig (check cfg.revision)
    -- => now done with rev 19!
    cfgversion = 12
  end

  if tonumber(cfgversion) == 12 then
    -- fix CustomLog directive in Apache
    local cfgfile
    if LC.fs.is_file("/etc/apache2/conf.d/liveconfig") then
      cfgfile = "/etc/apache2/conf.d/liveconfig"
    elseif LC.fs.is_file("/etc/apache2/conf.d/liveconfig.conf") then
      cfgfile = "/etc/apache2/conf.d/liveconfig.conf"
    elseif LC.fs.is_file("/etc/apache2/conf-available/liveconfig.conf") then
      cfgfile = "/etc/apache2/conf-available/liveconfig.conf"
    elseif LC.fs.is_file("/etc/apache2/modules.d/99_liveconfig.conf") then
      cfgfile = "/etc/apache2/modules.d/99_liveconfig.conf"
    elseif LC.fs.is_file("/etc/httpd/conf.d/99_liveconfig.conf") then
      cfgfile = "/etc/httpd/conf.d/99_liveconfig.conf"
    elseif LC.fs.is_file("/usr/local/etc/apache22/extra/liveconfig.conf") then
      cfgfile = "/usr/local/etc/apache22/extra/liveconfig.conf"
    end
    local cfg = LC.web.getConfig("apache")
    if cfgfile ~= nil and cfg ~= nil then
      local v_major, v_minor, v_patch = string.match(cfg["binversion"], "(%d+)%.(%d+)%.?(%d*)")
      v_major = tonumber(v_major)
      v_minor = tonumber(v_minor)
      v_patch = tonumber(v_patch)
      if v_major > 2 or (v_major == 2 and v_minor > 2) or (v_major == 2 and v_minor == 2 and v_patch >= 11) then
        LC.log.print(LC.log.INFO, "Patching CustomLog directive in ", cfgfile)
        local rc = LC.exec("sed -i -e 's/^CustomLog \"|exec \\//CustomLog \"||\\//g' " .. cfgfile)
      end
    end
    -- increase cfg version
    cfgversion = 13
  end

  if tonumber(cfgversion) == 13 or tonumber(cfgversion) == 14 or tonumber(cfgversion) == 15 then
    cfgversion = 16
  end

  if tonumber(cfgversion) == 16 then
    -- fix spamd config options:
    if LC.fs.is_file("/etc/spamassassin/liveconfig.status") or LC.fs.is_file("/etc/mail/spamassassin/liveconfig.status") then
      -- patch SpamAssassin start options
      if LC.distribution.family == "Debian" then
        LC.exec("sed -i -e 's/--socketowner=spamd --socketgroup=root --socketmode=0700/--socketowner=root --socketgroup=spamd --socketmode=0660/' /etc/default/spamassassin")
      elseif LC.distribution.family == "RedHat" then
        LC.exec("sed -i -e 's/^SPAMDOPTIONS=\"--max-children 5 --helper-home-dir /SPAMDOPTIONS=\"-d -m 5 -H /' -e 's/--socketowner=spamd --socketgroup=root --socketmode=0700/--socketowner=root --socketgroup=spamd --socketmode=0660/' /etc/sysconfig/spamassassin")
      elseif LC.distribution.family == "SUSE" then
        LC.exec("sed -i -e 's/^SPAMD_ARGS=\"--max-children 5 --helper-home-dir /SPAMD_ARGS=\"-d -m 5 -H /' -e 's/--socketowner=spamd --socketgroup=root --socketmode=0700/--socketowner=root --socketgroup=spamd --socketmode=0660/' /etc/sysconfig/spamd")
      elseif LC.distribution.family == "Gentoo" then
        LC.exec("sed -i -e 's/^SPAMD_OPTS=\"--max-children 5 --helper-home-dir /SPAMD_OPTS=\"-d -m 5 -H /' -e 's/--socketowner=spamd --socketgroup=root --socketmode=0700/--socketowner=root --socketgroup=spamd --socketmode=0660/' /etc/conf.d/spamd")
      end

      -- restarting lcsam with v27!
    end
    cfgversion = 17
  end

  if tonumber(cfgversion) == 17 or tonumber(cfgversion) == 18 then
    cfgversion = 19
  end

  if tonumber(cfgversion) == 19 then
    -- update lclogparse.conf and restart lclogparse:
    --> this is now done with #31
    cfgversion = 20
  end

  if tonumber(cfgversion) == 20 then
    if LC.fs.is_file("/etc/init.d/lcsam") and LC.fs.is_file("/usr/sbin/update-rc.d") and not LC.fs.is_file("/etc/rc3.d/S20lcsam") then
      -- fix missing "lcsam" init
      LC.exec("/usr/sbin/update-rc.d -f lcsam defaults")
    end
    cfgversion = 21
  end

  if tonumber(cfgversion) == 21 then
    local cfg = LC.web.getConfig("apache")
    if cfg ~= nil and cfg.revision ~= nil then
      local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
      local vhostfile = "default"
      if LC.fs.is_file(vhostpath .. "/000-default.conf") then
        vhostfile = "000-default.conf"
      elseif LC.fs.is_file(vhostpath .. "/00_default_vhost.conf") then
        vhostfile = "00_default_vhost.conf"
      end
      -- fix invalid RewriteRule in default vHost
      if LC.fs.is_file(vhostpath .. "/" .. vhostfile) then
        LC.exec("sed -i -e 's/RewriteRule ^\\$  \\[R=404,L\\]/RewriteRule ^\\$ - \\[R=404,L\\]/' " .. vhostpath .. "/" .. vhostfile)
        -- reload Apache configuration
        LC.timeout.set('apache.reload', 10, 60)
      end
    end
    cfgversion = 22
  end

  if tonumber(cfgversion) == 22 then
    -- fix "Options" directive in Apache configuration and .htaccess files
    local cfg = LC.web.getConfig("apache")
    if cfg ~= nil and cfg.revision ~= nil then
      LC.log.print(LC.log.INFO, "Running '", LC.liveconfig.libexecdir, "/lcservice.sh fix-options'")
      LC.exec(LC.liveconfig.libexecdir .. '/lcservice.sh fix-options')
      -- reload Apache configuration
      LC.timeout.set('apache.reload', 10, 60)
    end
    -- increase cfg version
    cfgversion = 23
  end

  if tonumber(cfgversion) == 23 then
    local cfg = LC.web.getConfig("apache")
    if cfg ~= nil and cfg.revision ~= nil then
      LC.log.print(LC.log.INFO, "Updating Apache default vHost configuration")
      local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
      local vhostfile
      if LC.distribution.family == "Debian" then
        if LC.fs.is_file(vhostpath .. "/000-default.conf") or LC.fs.is_file(vhostpath .. "/000-default.conf.lcbak") then
          vhostfile = "000-default.conf"
        else
          vhostfile = "default"
        end
      elseif LC.distribution.family == "Gentoo" or LC.distribution.family == "RedHat" then
        vhostfile = "00_default_vhost.conf"
      elseif LC.distribution.family == "SUSE" then
        vhostfile = "default.conf"
      end
      local datadir = string.gsub(LC.liveconfig.datadir, "/", "\\/")
      local rc = LC.exec("sed -i -e 's/^ServerName default/ServerName default\\nDocumentRoot \"" .. datadir .. "\\/html\\/\"/' " .. vhostpath .. "/" .. vhostfile)
      -- reload Apache configuration
      LC.timeout.set('apache.reload', 10, 60)
    end
    -- increase cfg version
    cfgversion = 24
  end

  if tonumber(cfgversion) == 24 then
    if LC.distribution.family == "Debian" then
      local cfg = LC.web.getConfig("apache")
      if cfg ~= nil and cfg.revision ~= nil and LC.fs.is_file("/etc/apache2/conf.d/liveconfig") then
        -- fix name of "liveconfig" configuration file for Apache httpd ("liveconfig" -> "liveconfig.conf")
        if LC.fs.is_file("/usr/sbin/a2enconf") and LC.fs.is_dir("/etc/apache2/conf-available") and not LC.fs.is_file("/etc/apache2/conf-available/liveconfig.conf") then
          -- Debian >=8
          LC.fs.rename("/etc/apache2/conf.d/liveconfig", "/etc/apache2/conf-available/liveconfig.conf")
          LC.exec("a2enconf liveconfig")
        elseif not LC.fs.is_file("/etc/apache2/conf.d/liveconfig.conf") then
          -- Debian <=7
          LC.fs.rename("/etc/apache2/conf.d/liveconfig", "/etc/apache2/conf.d/liveconfig.conf")
        end
        -- reload Apache configuration
        LC.timeout.set('apache.reload', 10, 60)
      end
    end
    -- increase cfg version
    cfgversion = 25
  end

  if tonumber(cfgversion) == 25 then
    local cfg = LC.dkim.detect()
    if cfg ~= nil and cfg.configured then
      -- DKIM is configured/enabled:
      -- create new opendkim socket directory if it doesn't exist
      if not LC.fs.is_dir("/var/spool/postfix/opendkim") then
        LC.log.print(LC.log.INFO, "Creating directory '/var/spool/postfix/opendkim'")
        LC.fs.mkdir("/var/spool/postfix/opendkim")
        -- adjust permissions
        LC.fs.setperm("/var/spool/postfix/opendkim", 750, "opendkim", "opendkim")
      end
      -- patch OpenDKIM configuration
      LC.log.print(LC.log.INFO, "Patching OpenDKIM configuration (new Socket)")
      LC.exec("sed -i -e 's/^\\(Socket.*\\)local:\\/var\\/run\\/opendkim\\/opendkim.sock/\\1local:\\/var\\/spool\\/postfix\\/opendkim\\/opendkim.sock/g' " .. cfg.configfile)
      LC.exec(cfg.restart_cmd)
      -- patch Postfix configuration to use new socket directory
      cfg = LC.smtp.getConfig("postfix")
      if cfg ~= nil and cfg.revision ~= nil then
        LC.log.print(LC.log.INFO, "Patching OpenDKIM socket in main.cf and restarting Postfix")
        LC.exec("sed -i -e 's/^\\(non_smtpd_milters = .*\\)unix:\\/var\\/run\\/opendkim\\/opendkim.sock\\(.*\\)/\\1unix:\\/opendkim\\/opendkim.sock\\2/' -e 's/^\\(smtpd_milters = .*\\)unix:\\/var\\/run\\/opendkim\\/opendkim.sock\\(.*\\)/\\1unix:\\/var\\/spool\\/postfix\\/opendkim\\/opendkim.sock\\2/g' " .. cfg.configfile)
        LC.exec(cfg.restart_cmd)
      end
    end
    -- increase cfg version
    cfgversion = 26
  end

  if tonumber(cfgversion) == 26 then
    -- replace "nginx-php-fcgi" symlink with real file (symlinks aren't supported any more
    -- at least on CentOS 7.2 with systemd)
    if LC.fs.is_symlink("/etc/init.d/nginx-php-fcgi") then
      LC.exec("rm -f /etc/init.d/nginx-php-fcgi")
      LC.exec("cp " .. LC.liveconfig.libexecdir .. "/nginx-php-fcgi /etc/init.d/")
    end
    -- same with "lcsam":
    if LC.fs.is_symlink("/etc/init.d/lcsam") then
      LC.exec("rm -f /etc/init.d/lcsam")
      LC.exec("cp " .. LC.liveconfig.libexecdir .. "/lcsam.init /etc/init.d/lcsam")
    end
    -- increase cfg version
    cfgversion = 27
  end

  if tonumber(cfgversion) == 27 then
    -- remove possibly existing catchall accounts from SpamAssassin user list
    if LC.fs.is_file("/etc/spamassassin/liveconfig.status") or LC.fs.is_file("/etc/mail/spamassassin/liveconfig.status") then
      if LC.fs.is_file("/etc/postfix/spamassassin") then
        LC.exec("sed -i -e '/^@.*/d' /etc/postfix/spamassassin")
        LC.spamassassin.update_map()
      end
      local cfg = LC.spamassassin.detect()
      if cfg ~= nil then
        -- restart lcsam
        if LC.fs.is_file("/sbin/service") then
          LC.exec("/sbin/service lcsam restart")
        else
          LC.exec("/etc/init.d/lcsam restart")
        end
      end
    end
    -- increase cfg version
    cfgversion = 28
  end

  if tonumber(cfgversion) == 28 then
    -- moved to v46
  end

  if tonumber(cfgversion) == 29 then
    local rc
    if LC.fs.is_dir("/etc/awstats/liveconfig") then
      -- fix "SkipHosts" option in AWStats
      rc=LC.exec("sed -i -e 's/^SkipHosts=\"0.0.0.0\"/SkipHosts=\"0.0.0.1\"/' /etc/awstats/liveconfig/awstats.*.conf")
    end
    if LC.fs.is_dir("/etc/webalizer") then
      -- fix "IgnoreSite" option in Webalizer
      rc=LC.exec("sed -i -e 's/^IgnoreSite      0.0.0.0/IgnoreSite      0.0.0.1/' /etc/webalizer/*.conf")
    end
    -- patch dummy entries in placeholder files
    local webroot = LC.web.getWebRoot()
    rc = LC.exec("find " .. webroot .. "/*/logs -maxdepth 1 -type f -name 'access.log.1' -size 90c -exec sed -i -e 's/^0.0.0.0 - - \\[01\\/Jan\\/2011/0.0.0.1 - - \\[01\\/Jan\\/2011/' {} \\;")
    -- increase cfg version
    cfgversion = 30
  end

  if tonumber(cfgversion) == 30 then
    -- fix owner of /var/lib/liveconfig
    LC.fs.setperm("/var/lib/liveconfig", "0755", "liveconfig", "liveconfig")
    -- increase cfg version
    cfgversion = 31
  end

  if tonumber(cfgversion) == 31 then
    local cfg = LC.smtp.getConfig("postfix")
    if cfg ~= nil and cfg.revision ~= nil then
      -- update lclogparse.conf and restart lclogparse
      -- -> moved to cfgversion=38!

      -- fix installation of systemd unit file for "lcsam":
      if LC.fs.is_file("/etc/init.d/lcsam") and LC.fs.is_dir("/lib/systemd/system") and not LC.fs.is_file("/lib/systemd/system/lcsam.service") then
        LC.exec("cp --preserve=all " .. LC.liveconfig.libexecdir .. "/lcsam.service /lib/systemd/system/")
        if LC.fs.is_file("/bin/systemctl") then
          LC.exec("/bin/systemctl enable lcsam")
        end
      end

      -- start lcsam (if not running)
      if LC.fs.is_file("/sbin/service") then
        LC.exec("/sbin/service lcsam start")
      else
        LC.exec("/etc/init.d/lcsam start")
      end
    end
    cfgversion = 32
  end

  if tonumber(cfgversion) == 32 then
    -- patch dovecot.conf (add "passdb" section with "deny = yes")
    local cfg = LC.popimap.getConfig("dovecot")
    if cfg ~= nil and cfg.revision ~= nil then
      local v_major, v_minor = string.match(cfg.binversion, "^(%d+)%.(%d+)")
      v_major = tonumber(v_major)
      v_minor = tonumber(v_minor)
      if v_major == 1 then
        -- Dovecot version 1.x
        LC.log.print(LC.log.INFO, "Updating ", cfg["configfile"], " (Dovecot 1.x)")
        LC.exec("grep -q 'deny = yes' " .. cfg["configfile"] .. " || sed -i -e 's/^passdb passwd-file {$/passdb passwd-file {\\n  args = \\/etc\\/dovecot\\/deny.%s\\n  deny = yes\\n}\\npassdb passwd-file {/' " .. cfg["configfile"])
      else
        -- Dovecot version 2.x or later
        LC.log.print(LC.log.INFO, "Updating ", cfg["configfile"], " (Dovecot 2.x)")
        LC.exec("grep -q 'deny = yes' " .. cfg["configfile"] .. " || sed -i -e 's/^passdb {$/passdb {\\n  args = \\/etc\\/dovecot\\/deny.%s\\n  driver = passwd-file\\n  deny = yes\\n}\\npassdb {/' " .. cfg["configfile"])
      end
    end
    cfgversion = 33
  end

  if tonumber(cfgversion) == 33 then
    -- create deny.imap|pop3|imap|lda if not existing
    local cfg = LC.popimap.getConfig("dovecot")
    if cfg ~= nil and cfg.revision ~= nil then
      if not LC.fs.is_file(cfg["configpath"] .. "/deny.imap") then
        local fhr, msg = io.open(cfg["configpath"] .. "/deny.imap", "a")
        if fhr == nil then
          LC.log.print(LC.log.ERR, "Can't open '", cfg["configpath"], "/deny.imap' for appending: ", msg)
          return false, "Can't open '" .. cfg["configpath"] .. "/deny.imap' for appending: " .. msg
        end
        -- adjust owner & permissions - only Dovecot needs access to this file, so set mod 600 !!!
        LC.fs.setperm(cfg["configpath"] .. "/deny.imap", "0600", "dovecot", "root")
        fhr:close()
      end
      if not LC.fs.is_file(cfg["configpath"] .. "/deny.pop3") then
        -- create symlink from deny.pop3 to deny.imap:
        os.execute("ln -s " .. cfg["configpath"] .. "/deny.imap " .. cfg["configpath"] .. "/deny.pop3")
      end
      if not LC.fs.is_file(cfg["configpath"] .. "/deny.smtp") then
        local fhr, msg = io.open(cfg["configpath"] .. "/deny.smtp", "a")
        if fhr == nil then
          LC.log.print(LC.log.ERR, "Can't open '", cfg["configpath"], "/deny.smtp' for appending: ", msg)
          return false, "Can't open '" .. cfg["configpath"] .. "/deny.smtp' for appending: " .. msg
        end
        -- adjust owner & permissions - only Dovecot needs access to this file, so set mod 600 !!!
        LC.fs.setperm(cfg["configpath"] .. "/deny.smtp", "0600", "dovecot", "root")
        fhr:close()
      end
      if not LC.fs.is_file(cfg["configpath"] .. "/deny.lda") then
        -- create symlink from deny.pop3 to deny.imap:
        os.execute("ln -s " .. cfg["configpath"] .. "/deny.smtp " .. cfg["configpath"] .. "/deny.lda")
      end

      -- reload Dovecot configuration
      LC.exec(cfg["reload_cmd"])
    end
    cfgversion = 34
  end

  if tonumber(cfgversion) == 34 then
    -- install new GPG repository key on RPM based systems
    if LC.distribution.family == "RedHat" or LC.distribution.family == "SUSE" then
      LC.exec("rpm -qi gpg-pubkey-3a2b2840-59c214b7 || rpm --import /etc/pki/rpm-gpg/RPM-GPG-KEY-LiveConfig")
    end
    cfgversion = 35
  end

  if tonumber(cfgversion) == 35 then
    local cfg = LC.web.getConfig("apache")
    if cfg ~= nil and cfg.revision ~= nil then
      local v_major, v_minor, v_patch = string.match(cfg.binversion, "^(%d+)%.(%d+)%.(%d+)")
      v_major = tonumber(v_major)
      v_minor = tonumber(v_minor)
      v_patch = tonumber(v_patch)
      if v_major > 2 or (v_major == 2 and (v_minor > 3 or (v_minor == 3 and v_patch >= 3))) then
        -- OCSP supported since v2.3.3
        local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
        local vhostfile
        if LC.distribution.family == "Debian" then
          if LC.fs.is_file(vhostpath .. "/000-default.conf") or LC.fs.is_file(vhostpath .. "/000-default.conf.lcbak") then
            vhostfile = "000-default.conf"
          else
            vhostfile = "default"
          end
        elseif LC.distribution.family == "Gentoo" or LC.distribution.family == "RedHat" then
          vhostfile = "00_default_vhost.conf"
        elseif LC.distribution.family == "SUSE" then
          vhostfile = "default.conf"
        end
        vhostfile = vhostpath .. "/" .. vhostfile
        LC.log.print(LC.log.INFO, "Updating Apache default vHost configuration (OCSP settings)")
        local run_dir = string.gsub(cfg.run_dir, "/", "\\/")
        local rc = LC.exec("grep -q SSLStapling " .. vhostfile .. " || sed -i -e 's/^\\(  SSLCipherSuite .*\\)/\\1\\n  SSLStaplingReturnResponderErrors off\\n  SSLStaplingResponderTimeout 5\\n  SSLStaplingCache shmcb:" .. run_dir .. "\\/ssl_stapling_cache(128000)/' " .. vhostfile)
        -- reload Apache configuration
        LC.timeout.set('apache.reload', 10, 60)
      end
    end
    -- increase cfg version
    cfgversion = 36
  end

  if tonumber(cfgversion) == 36 then
    local cfg = LC.web.getConfig("nginx")
    if cfg ~= nil and cfg.revision ~= nil then
      -- create /etc/nginx/conf.d/resolver.conf if not existing
      if not LC.fs.is_file(cfg["configpath"] .. "/conf.d/resolver.conf") then
        nginx.updateResolver()
      end
    end
    -- increase cfg version
    cfgversion = 37
  end

  if tonumber(cfgversion) == 37 then
    local cfg = LC.web.getConfig("apache")
    if cfg ~= nil and cfg.revision ~= nil and LC.fs.is_file(cfg.configpath .. "/accesslog.map") then
      -- clean up Apache accesslog.map
      LC.web.cleanupAccesslogMap(cfg.configpath .. "/accesslog.map")
    end
    cfg = LC.web.getConfig("nginx")
    if cfg ~= nil and cfg.revision ~= nil and LC.fs.is_file(cfg.configpath .. "/accesslog.map") then
      -- clean up NGINX accesslog.map
      LC.web.cleanupAccesslogMap(cfg.configpath .. "/accesslog.map")
    end
    -- increase cfg version
    cfgversion = 38
  end

  if tonumber(cfgversion) == 38 then
    -- update lclogparse.conf and restart lclogparse:
    --> this is now done with #44
  end

  if tonumber(cfgversion) == 39 then

    local cfg = LC.web.getConfig("nginx")
    if cfg ~= nil and cfg.revision ~= nil then
      local v_major, v_minor, v_patch = string.match(cfg.binversion, "^(%d+)%.(%d+)%.(%d+)")
      v_major = tonumber(v_major)
      v_minor = tonumber(v_minor)
      v_patch = tonumber(v_patch)
      if LC.fs.is_file(cfg["configpath"] .. "/conf.d/resolver.conf") and (v_major < 1 or (v_major == 1 and (v_minor < 2 or (v_minor == 2 and v_patch < 2) or (v_minor == 3 and v_patch < 1)))) then
        -- fix configuration for NGINX < 1.2.2 and < 1.3.1 (eg. on Debian 7) when using IPv6 resolvers
        nginx.updateResolver()
      end
    end
    -- increase cfg version
    cfgversion = 40
  end

  if tonumber(cfgversion) == 40 then
    local cfg = LC.web.getConfig("apache")
    if cfg ~= nil and LC.fs.is_file(cfg.configfile) then
      LC.log.print(LC.log.INFO, "Patching CustomLog directive in ", cfg.configfile)
      local rc = LC.exec("sed -i -e 's/^CustomLog.*$/CustomLog \"||\\/usr\\/lib\\/liveconfig\\/lclogsplit -i -w\" LiveConfig/' " .. cfg.configfile)
      -- reload Apache configuration
      LC.timeout.set('apache.reload', 10, 60)
    end

    cfg = LC.web.getConfig("nginx")
    if cfg ~= nil and LC.fs.is_file(cfg.configfile) then
      LC.log.print(LC.log.INFO, "Patching access_log directive in ", cfg.configfile)
      local rc = LC.exec("grep -q '^access_log' " .. cfg.configfile .. " || sed -i -e \"s/\\(bytes_sent';\\)/\\1\\n\\naccess_log \\/var\\/log\\/nginx\\/access.liveconfig.log LiveConfig;/\" " .. cfg.configfile)
      -- reload NGINX configuration
      LC.timeout.set('nginx.reload', 10, 60)

      -- remove old /etc/nginx/accesslog.map:
      os.execute("rm -f /etc/nginx/accesslog.map")

      -- install and enable service "lclogsplit":
      nginx.lclogsplit_enable(cfg)

    end

    -- increase cfg version
    cfgversion = 41
  end

  if tonumber(cfgversion) == 41 then
    local cfg = LC.smtp.getConfig("postfix")
    if cfg ~= nil and cfg.revision ~= nil and postfix.LOCALMASTER["dovecot"] == nil then
      -- patch/add "null_sender" option in /etc/postfix/master.cf
      local masterfile = cfg["configpath"] .. "/master.cf"
      LC.log.print(LC.log.INFO, "Patching/adding 'null_sender' option for dovecot delivery in ", masterfile)
      local dovecot_deliver = "/usr/lib/dovecot/deliver"
      if LC.fs.is_file("/usr/libexec/dovecot/deliver") then
        dovecot_deliver = "/usr/libexec/dovecot/deliver"
      end
      dovecot_deliver = string.gsub(dovecot_deliver, "/", "\\/")
      LC.exec("sed -i -e 's/flags=DROhu user=mail:mail \\(null_sender=.*\\)\\?argv=" .. dovecot_deliver .. "/flags=DROhu user=mail:mail null_sender= argv=" .. dovecot_deliver .. "/' " .. masterfile)

      -- restart Postfix:
      LC.log.print(LC.log.INFO, "Restarting Postfix service")
      LC.exec(cfg.restart_cmd)
    end

    cfg = LC.popimap.getConfig("dovecot")
    if cfg ~= nil and cfg.revision ~= nil then
      if not LC.fs.is_file(cfg["configpath"] .. "/deny.sieve") and LC.fs.is_file(cfg["configpath"] .. "/deny.imap") then
        -- create symlink from deny.sieve to deny.imap:
        os.execute("ln -s " .. cfg["configpath"] .. "/deny.imap " .. cfg["configpath"] .. "/deny.sieve")
      end
    end

    -- increase cfg version
    cfgversion = 42
  end

  if tonumber(cfgversion) == 42 then
    local cfg = LC.ftp.getConfig("proftpd")
    if cfg ~= nil and cfg.revision ~= nil then
      local v_major, v_minor, v_patch = string.match(cfg["binversion"], "(%d+)%.(%d+)%.?(%d*)")
      v_major = tonumber(v_major)
      v_minor = tonumber(v_minor)
      v_patch = tonumber(v_patch)

      if v_major > 1 or (v_major == 1 and v_minor > 3) or (v_major == 1 and v_minor == 3 and v_patch >= 5) then
        -- patch "TLSProtocol" setting to also support TLS 1.1 and 1.2
        LC.exec("sed -i -e 's/^  TLSProtocol TLSv1$/  TLSProtocol TLSv1 TLSv1.1 TLSv1.2/' " .. cfg.configfile)

        -- reload ProFTPD configuration:
        LC.log.print(LC.log.INFO, "Reloading ProFTPD configuration")
        LC.exec(cfg.reload_cmd)
      end
    end

    -- increase cfg version
    cfgversion = 43
  end

  if tonumber(cfgversion) == 43 then
    local cfg_apache = LC.web.getConfig("apache")
    local cfg_nginx = LC.web.getConfig("nginx")
    if (cfg_apache ~= nil and cfg_apache.revision ~= nil) or (cfg_nginx ~= nil and cfg_nginx.revision ~= nil) then
      local webroot = LC.web.getWebRoot()
      if LC.fs.is_dir(webroot) then
        LC.log.print(LC.log.INFO, "Setting immutable flag for PHP directories below ", webroot, "/*/conf/")
        LC.exec("find " .. webroot .. "/*/conf -maxdepth 2 -type f -name \"php.ini\" -exec sh -c 'name=`dirname \"$0\"`; chattr -f +i \"$name\"; ' {} \\;")
      end
    end

    -- increase cfg version
    cfgversion = 44
  end

  if tonumber(cfgversion) == 44 then
    local cfg = LC.smtp.getConfig("postfix")
    if cfg ~= nil and cfg.revision ~= nil then
      if LC.fs.is_file(LC.liveconfig.sysconfdir .. "/lclogparse.conf") then
        -- update lclogparse.conf and restart lclogparse:
        postfix.update_lclogparse(cfg)
      end
    end
    cfgversion = 45
  end

  if tonumber(cfgversion) == 45 then
    -- remove ~/conf/acme from all webspaces (not required any more)
    local webroot = LC.web.getWebRoot()
    LC.exec("rm -rf " .. webroot .. "/*/conf/acme")
    cfgversion = 46
  end

  if tonumber(cfgversion) == 46 then
    local cfg = LC.web.getConfig("apache")
    if cfg ~= nil and LC.fs.is_file(cfg.configfile) then
      LC.log.print(LC.log.INFO, "Patching LogFormat directive in ", cfg.configfile)
      local rc = LC.exec("sed -i -e 's/^  LogFormat \"%%v:#:%%h/  LogFormat \"%v:#:%a/' " .. cfg.configfile)
    end
    -- increase cfg version
    cfgversion = 47
  end

  if tonumber(cfgversion) == 47 then
    -- patch Postfix configuration to add "local_recipient_maps"
    local cfg = LC.smtp.getConfig("postfix")
    if cfg ~= nil and cfg.revision ~= nil then
      LC.log.print(LC.log.INFO, "Adding local_recipient_maps to Postfix main.cf")
      LC.exec("grep -q '^local_recipient_maps' " .. cfg.configfile .. " || sed -i -e 's/^\\(# <EOF>\\)/local_recipient_maps = $alias_maps\\n\\1/' " .. cfg.configfile)
      LC.exec(cfg.restart_cmd)
    end
    -- increase cfg version
    cfgversion = 48
  end

  if tonumber(cfgversion) == 48 then
    -- patch logrotate-vhosts
    if LC.fs.is_file("/etc/logrotate.d/liveconfig-vhosts") then
      LC.log.print(LC.log.INFO, "Patching /etc/logrotate.d/liveconfig-vhosts")
      local rc = LC.exec("sed -i -e 's/^    su [^ ]\\+/    su root/' /etc/logrotate.d/liveconfig-vhosts")
    end
    -- increase cfg version
    cfgversion = 49
  end

  if tonumber(cfgversion) == 49 then
    -- Postfix: remove invalid records from sender_bcc / recipient_bcc
    local cfg = LC.smtp.getConfig("postfix")
    if cfg ~= nil and cfg.revision ~= nil then
      local bcc_file = cfg.configpath .. "/recipient_bcc"
      if LC.fs.is_file(bcc_file) then
        LC.log.print(LC.log.INFO, "Cleaning Postfix recipient_bcc")
        LC.exec("sed -i -e 's/^[^\\t]*\\t$//' " .. bcc_file)
        LC.exec("/usr/sbin/postmap " .. bcc_file)
      end
      bcc_file = cfg.configpath .. "/sender_bcc"
      if LC.fs.is_file(bcc_file) then
        LC.log.print(LC.log.INFO, "Cleaning Postfix sender_bcc")
        LC.exec("sed -i -e 's/^[^\\t]*\\t$//' " .. bcc_file)
        LC.exec("/usr/sbin/postmap " .. bcc_file)
      end
    end
    -- increase cfg version
    cfgversion = 50
  end

  if tonumber(cfgversion) == 50 then
    cfgversion = 51
  end

  if tonumber(cfgversion) == 51 then
    if LC.distribution.family == "Debian" then
      -- Debian: move chain certificates from /etc/ssl/certs/*-ca.crt into /etc/ssl/chains/*-ca.crt
      -- (local validation problem, possible bug in OpenSSL)
      -- Then run "update-ca-certificates" to remove all old hash links
      -- additionally, update Apache2 and ProFTPD configurations accordingly
      local cfg = LC.web.getConfig("apache")
      if cfg ~= nil and cfg.revision ~= nil and cfg.ssl_chain_dir ~= nil and cfg.ssl_chain_dir ~= cfg.ssl_cert_dir then
        if not LC.fs.is_dir(cfg.ssl_chain_dir) then
          LC.log.print(LC.log.INFO, "Creating directory for TLS/SSL chain certificates: ", cfg.ssl_chain_dir)
          LC.fs.mkdir(cfg.ssl_chain_dir)
          LC.fs.setperm(cfg.ssl_chain_dir, 755, "root", "root")
        end
        -- move all existing chain files:
        LC.log.print(LC.log.INFO, "Moving TLS/SSL chain files to ", cfg.ssl_chain_dir)
        local glob = '[0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f]'
        LC.exec("/bin/mv " .. cfg.ssl_cert_dir .. "/" .. glob .. "-ca.crt " .. cfg.ssl_chain_dir .. "/")
        LC.exec("/bin/mv " .. cfg.ssl_cert_dir .. "/httpd-default-ca.crt " .. cfg.ssl_chain_dir .. "/")
        -- fix Apache vHosts
        LC.exec("sed -i --follow-symlinks -e 's/^\\s*SSLCertificateChainFile \\/etc\\/ssl\\/certs\\//    SSLCertificateChainFile \\/etc\\/ssl\\/chains\\//' " .. cfg["enabled_dir"] .. "/*")
        -- reload Apache configuration
        LC.timeout.set('apache.reload', 10, 60)
      end

      -- adjust ProFTPD
      cfg = LC.ftp.getConfig("proftpd")
      if cfg ~= nil and cfg.revision ~= nil and cfg.cafile ~= nil then
        -- also fix ProFTPD configuration
        local cadir = LC.fs.dirname(cfg.cafile)
        if not LC.fs.is_dir(cadir) then
          LC.log.print(LC.log.INFO, "Creating directory for TLS/SSL chain certificates: ", cadir)
          LC.fs.mkdir(cadir)
          LC.fs.setperm(cadir, 755, "root", "root")
        end
        -- move existing chain file:
        LC.log.print(LC.log.INFO, "Moving TLS/SSL chain file to ", cfg.cafile)
        LC.exec("/bin/mv /etc/ssl/certs/proftpd-ca.crt " .. cfg.cafile)
        -- fix ProFTPD configuration
        LC.exec("sed -i --follow-symlinks -e 's/^  TLSCertificateChainFile \\/etc\\/ssl\\/certs\\/proftpd-ca.crt/  TLSCertificateChainFile \\/etc\\/ssl\\/chains\\/proftpd-ca.crt/' /etc/proftpd/proftpd.conf")
        -- reload Apache configuration
        LC.timeout.set('proftpd.reload', 10, 60)
      end

      -- rehash - this also removes all old hash symlinks
      LC.exec("/usr/bin/c_rehash 1>/dev/null 2>&1")
    end

    -- increase cfg version
    cfgversion = 52
  end

  if tonumber(cfgversion) == 52 then
    local cfg = LC.dns.getConfig("bind")
    if cfg ~= nil and cfg.revision ~= nil then
      -- improve permissions of zone file directories
      if LC.fs.is_dir(cfg.zonepath) then
        LC.log.print(LC.log.INFO, "Adjusting file permissions: ", cfg.zonepath)
        LC.fs.setperm(cfg.zonepath, "0770", cfg.cfgfileowner, cfg.cfgfilegroup)
      end
      if LC.fs.is_dir(cfg.cachepath) then
        LC.log.print(LC.log.INFO, "Adjusting file permissions: ", cfg.cachepath)
        LC.fs.setperm(cfg.cachepath, "0770", cfg.cfgfileowner, cfg.cfgfilegroup)
      end

    end
    cfgversion = 53
  end

  if tonumber(cfgversion) == 53 then
    -- moved to r54
    cfgversion = 54
  end

  if tonumber(cfgversion) == 54 then
    -- moved to r56
    cfgversion = 55
  end

  if tonumber(cfgversion) == 55 then
    if LC.fs.is_dir("/etc/webalizer") then
      -- fix "IgnoreSite" option in Webalizer
      rc=LC.exec("grep -L IncrementalName /etc/webalizer/*.conf | while read fname; do sed -i -e 's/^Incremental     yes/Incremental     yes\\nIncrementalName webalizer.current\\nHistoryName     webalizer.hist\\nDNSCache        dns_cache.db/' $fname ; done")
    end
    cfgversion = 56
  end

  if tonumber(cfgversion) == 56 then
    -- moved to r57
    cfgversion = 57
  end

  if tonumber(cfgversion) == 57 then
    -- patch dovecot.conf (add "stats-writer" directive for Dovecot 2.3+)
    local cfg = LC.popimap.getConfig("dovecot")
    if cfg ~= nil and cfg.revision ~= nil then
      local v_major, v_minor = string.match(cfg.binversion, "^(%d+)%.(%d+)")
      v_major = tonumber(v_major)
      v_minor = tonumber(v_minor)
      if v_major > 2 or (v_major == 2 and v_minor >= 3) then
        -- Dovecot version 2.3 or later
        -- fix wrong escaping from r56
        LC.exec("sed -i -e 's/\\\\n/\\n/g' -e 's/^-e//g' " .. cfg["configfile"])
        -- otherwise add
        LC.log.print(LC.log.INFO, "Updating ", cfg["configfile"], " (Dovecot 2.3+)")
        LC.exec("grep -q 'stats-writer' " .. cfg["configfile"] .. " || sed -i -e 's/^\\(# <EOF>.*\\)$/\\1\\nservice stats {\\n  unix_listener stats-writer {\\n     user = mail\\n     group = mail\\n  }\\n}\\n/' " .. cfg["configfile"])
      end
    end
    cfgversion = 58
  end

  if tonumber(cfgversion) == 58 then
    local cfg = LC.web.getConfig("apache")
    if cfg ~= nil and LC.fs.is_file(cfg.configfile) then
      local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
      LC.log.print(LC.log.INFO, "Disabling MultiViews option for Apache cgi-bin directories in " .. vhostpath .. ":")
      local rc = LC.exec("grep -L 'Options -MultiViews' " .. vhostpath .. "/*.conf | while read fname; do grep -q 'ScriptAlias' \"$fname\" && echo -n \" $fname\" && sed -i --follow-symlinks -e 's/\\(ScriptAlias \\/cgi-bin\\/ \\(.*\\)\\)$/\\1\\n    <Directory \\2>\\n        Options -MultiViews\\n    <\\/Directory>/' \"$fname\"; done; echo \"\"")
      -- reload Apache configuration
      LC.timeout.set('apache.reload', 10, 60)
    end
    -- increase cfg version
    cfgversion = 59
  end

  if tonumber(cfgversion) == 59 then
    -- add workaround against SMTP smuggling (https://www.postfix.org/smtp-smuggling.html)
    local cfg = LC.smtp.getConfig("postfix")
    if cfg ~= nil and cfg.revision ~= nil then
      LC.log.print(LC.log.INFO, "Adding smtpd_data_restrictions to Postfix main.cf")
      LC.exec("grep -q '^smtpd_data_restrictions' " .. cfg.configfile .. " || sed -i -e 's/^\\(# <EOF>\\)/smtpd_data_restrictions = reject_unauth_pipelining\\n\\1/' " .. cfg.configfile)
      LC.exec(cfg.restart_cmd)
    end
    -- increase cfg version
    cfgversion = 60
  end

  -- write new sysconfig file:
  fh, errmsg = io.open(LC.liveconfig.sysconfdir .. "/sysconfig", "w")
  if fh == nil then
    LC.log.print(LC.log.INFO, "Can't open '" .. LC.liveconfig.sysconfdir .. "/sysconfig' for writing: ", errmsg)
    return false
  end

  -- write header
  LC.liveconfig.writeHeader(fh)

  -- write information
  fh:write([[
# This file is automatically created and updated by LiveConfig.
# DO NOT MAKE ANY CHANGES HERE!

]])

  -- write variables:
  fh:write("LC_WEBROOT=", LC.web.getWebRoot(), "\n")
  fh:write("LC_CFGVERSION=", cfgversion, "\n")

  -- write footer
  fh:write("\n")
  LC.liveconfig.writeFooter(fh)

  fh:close()
  return true
end

-- ---------------------------------------------------------------------------
-- LC.liveconfig.update_lclogparse()
-- Update lclogparse(1) configuration in /etc/liveconfig/lclogparse.cfg
-- ---------------------------------------------------------------------------
function LC.liveconfig.update_lclogparse(logfile, rules)
  LC.log.print(LC.log.DEBUG, "update_lclogparse(", logfile, ")")

  -- create lclogparse.conf
  local fh, errmsg = io.open(LC.liveconfig.sysconfdir .. "/lclogparse.conf", "w")
  if fh == nil then
    LC.log.print(LC.log.INFO, "Can't open '" .. LC.liveconfig.sysconfdir .. "/lclogparse.conf' for writing: ", errmsg)
    return false
  end

  LC.liveconfig.writeHeader(fh)

  fh:write("!STATUS-FILE:" .. LC.liveconfig.localstatedir .. "/lib/liveconfig/lclogparse.status\n")
  fh:write("!PID-FILE:" .. LC.liveconfig.runstatedir .. "/lclogparse.pid\n")

  fh:write("\n", logfile, "\n")
  fh:write(rules, "\n")

  LC.liveconfig.writeFooter(fh)
  fh:close()

  -- install service (systemd)
  if LC.fs.is_dir("/lib/systemd/system") and not LC.fs.is_file("/lib/systemd/system/lclogparse.service") then
    LC.exec("cp --preserve=all " .. LC.liveconfig.libexecdir .. "/lclogparse.service /lib/systemd/system/")
    if LC.fs.is_file("/bin/systemctl") then
      LC.exec("/bin/systemctl enable lclogparse")
    end
  end

  -- install init script
  if not LC.fs.is_file("/etc/init.d/lclogparse") then
    LC.exec("cp " .. LC.liveconfig.libexecdir .. "/lclogparse.init /etc/init.d/lclogparse")
    LC.fs.setperm("/etc/init.d/lclogparse", "0700", "root", "root")

    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d lclogparse defaults")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --add lclogparse")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update add lclogparse default")
    else
      LC.log.print(LC.log.ERR, "Can't install init script 'lclogparse' - no supported tool found")
    end
  end

  -- restart service
  if LC.fs.is_file("/sbin/service") then
    LC.exec("/sbin/service lclogparse restart")
  else
    LC.exec("/etc/init.d/lclogparse restart")
  end

  return true

end

-- ---------------------------------------------------------------------------
-- load_lua:d()
-- Load Lua configuration files from /etc/liveconfig/lua.d/*.lua
-- ---------------------------------------------------------------------------
local function load_lua_d()
  local files, errmsg, errno = LC.fs.glob(LC.liveconfig.sysconfdir .. "/lua.d/*.lua")
  if files == nil then
    LC.log.print(LC.log.ERR, "glob(" .. LC.liveconfig.sysconfdir .. "/lua.d/*.lua) failed: " .. errmsg .. " (" .. errno .. ")")
    return
  end

  local key, filename, f
  for key, filename in pairs(files) do
    LC.log.print(LC.log.DEBUG, "Loading '" .. filename .. "'")
    f, errmsg = loadfile(filename)
    if f == nil then
      LC.log.print(LC.log.ERR, "Error while loading '" .. filename .. "': " .. errmsg)
    end
    -- run parsed code:
    f()
  end
end

-- ---------------------------------------------------------------------------
-- entry point ("main()") in liveconfig.lua:
-- ---------------------------------------------------------------------------

-- load /etc/liveconfig/lua.d/*.lua
load_lua_d()

-- If "custom.lua" exists, load it:
if (LC.fs.is_file(LC.liveconfig.lua_path .. "/custom.lua")) then
  LC.log.print(LC.log.INFO, "Loading custom Lua settings from '", LC.liveconfig.lua_path, "/custom.lua'")
  dofile(LC.liveconfig.lua_path .. "/custom.lua")
end

if LC.liveconfig.DEFAULT_SSL_PCI_CIPHERS ~= nil then
  LC.log.print(LC.log.WARNING, "LC.liveconfig.SSL_PCI_CIPHERS is deprecated - use LC.liveconfig.SSL_STRONG_CIPHERS instead")
  LC.liveconfig.DEFAULT_SSL_STRONG_CIPHERS = LC.liveconfig.DEFAULT_SSL_PCI_CIPHERS
end

-- Detect os/distribution:
LC.distribution.detect()
LC.web.detect()
-- Done.

-- <EOF>----------------------------------------------------------------------
