--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2022 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/nginx.lua
-- Lua module to manage NGINX web server
-- ---------------------------------------------------------------------------
-- This is the "driver" module for NGINX web server.
-- It must be loaded by the command
--   LC.web.load("nginx")
-- Usually, this should happen at liveconfig.lua (or, if you have a customized
-- module, at custom.lua)
-- ---------------------------------------------------------------------------

-- Imported functions
local LC = LC
local LCS = LCS -- import liveconfig global storage
local type = type
local io = io
local os = os
local ipairs = ipairs
local pairs = pairs
local string = string
local table = table     -- for table.sort
local tonumber = tonumber

local print = print

-- Module declaration
module("nginx")

-- Define behaviour of 'coming soon' page (via 403 error):
-- 0: do not create a "location = /" section for "coming soon" pages
-- 1: [DEFAULT] create a "location = /" block to show "coming coon" page on any empty (sub)domain
-- 2: show "coming soon" page only on empty (sub)domains configured in webspace root directory (/htdocs/)
if COMING_SOON == nil then
  COMING_SOON = 1
end

-- default proxy parameters:
if PROXY_PARAMS == nil then
  PROXY_PARAMS = {
    ["proxy_redirect"] = "off",
    ["proxy_set_header"] = {
      "Host $host",
      "X-Real-IP $remote_addr",
      "X-Forwarded-For $proxy_add_x_forwarded_for",
      "X-Forwarded-Proto $scheme"
    },
    ["proxy_pass_header"] = {
      "Set-Cookie"
    },
    ["proxy_hide_header"] = {
      "Upgrade"
    },
    ["client_max_body_size"] = "64m",
    ["client_body_buffer_size"] = "128k",
    ["client_body_temp_path"] = "/var/cache/nginx/tmp"
  }
end

-- Exported functions
--   detect()
--   install()
--   uninstall()
--   configure()
--   configureVHost()
--   reload()
--   deleteAccount()
--   updateResolver()

-- do a "restart" instead of a "reload":
local flag_restart = false

-- ---------------------------------------------------------------------------
-- escape()
--
-- Escape quotes in a string to make them safe for config files
-- ---------------------------------------------------------------------------
local function escape(s)
  s=string.gsub(s, '\\', '\\\\')
  s=string.gsub(s, '"', '\\"')
  return(s)
end

-- ---------------------------------------------------------------------------
-- escapePCRE()
--
-- Escape special chars in a string to make them safe for NGINX regular expressions
-- ---------------------------------------------------------------------------
local function escapePCRE(s)
  s = string.gsub(s, "([%^%$%(%)%.%[%]%*%.%?|\\])", "\\%1")
  return s
end

-- ---------------------------------------------------------------------------
-- normalizeSlashes()
--
-- Replace "//" by "/"
-- ---------------------------------------------------------------------------
local function normalizeSlashes(s)
  while string.find(s, '//') ~= nil do
    s = string.gsub(s, '//', '/')
  end
  return s
end

-- ---------------------------------------------------------------------------
-- getBinaryVersion()
--
-- Return the binary version number from NGINX
-- ---------------------------------------------------------------------------
local function getBinaryVersion(bin)
  local handle = io.popen(bin .. " -v 2>&1", "r")
  local ret = handle:read("*a")
  handle:close()
  local v = string.match(ret, "nginx/(%d+%.%d+[%.%d+]*)")
  return v
end

-- ---------------------------------------------------------------------------
-- check_http2()
--
-- Check if NGINX version supports HTTP/2
-- ---------------------------------------------------------------------------
local function check_http2(data)
  local v_major, v_minor, v_patch = string.match(data.binversion, "^(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)
  if v_major > 1 or (v_major == 1 and (v_minor > 9 or (v_minor == 9 and v_patch >= 5))) then
    data.has_http2 = true
  else
    data.has_http2 = false
  end
end

-- ---------------------------------------------------------------------------
-- detect()
--
-- Detect if NGINX is installed. If yes, return a table containing
-- all relevant informations:
--   type          => "nginx"
--   binname       => file name of the NGINX binary (eg. "/usr/sbin/nginx")
--   binversion    => binary version (eg. "0.6.32")
--   pkgname       => name of the package (see below)
--   pkgversion    => package version number (eg. "0.6.32-3+lenny3")
--   configpath    => Main configuration path for NGINX
--   configfile    => LiveConfigs' configuration file (snippet)
--   statusfile    => LiveConfigs' status file (containing config version etc.)
--   start_cmd     => command to start NGINX
--   stop_cmd      => command to stop NGINX
--   reload_cmd    => command to re-load configuration
--   restart_cmd   => command to re-start NGINX process
--   available_dir => path for all available vhost configuration files
--   enabled_dir   => path for enabled vhost configuration files
--   enable_cmd    => command to enable a vhost configuration file (eg. "/usr/sbin/a2ensite")
--   disable_cmd   => command to disable a vhost configuration file (eg. "/usr/sbin/a2dissite")
--   httpd_user    => user running NGINX
--   httpd_group   => group running NGINX
--   has_http2     => 'true' if NGINX version supports HTTP/2 (>= 1.9.5)
--
-- If pkgversion is 'nil', then NGINX was most propably not installed
-- as native distribution package, but compiled/installed manually.
-- ---------------------------------------------------------------------------
function detect()

  -- first check for distribution-specific packages
  if LC.distribution.family == "Debian" then
    -- Debian/Ubuntu
    -- Package is named "nginx"
    local pkg, v = LC.distribution.hasPackage('nginx', 'nginx-full', 'nginx-light', 'nginx-naxsi', 'nginx-extras')
    if pkg ~= nil then
      LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
      local bin = "/usr/sbin/nginx"
      -- get binary version
      local bv = getBinaryVersion(bin)
      LC.log.print(LC.log.DEBUG, "binver ", bv)
      if bv ~= nil then
        -- ok, we have all informations. return data table:
        local data = {
          ["type"]          = "nginx",
          ["binname"]       = bin,
          ["binversion"]    = bv,
          ["pkgname"]       = pkg,
          ["pkgversion"]    = v,
          ["configpath"]    = "/etc/nginx",
          ["configfile"]    = "/etc/nginx/conf.d/liveconfig.conf",
          ["statusfile"]    = "/etc/nginx/liveconfig.status",
          ["start_cmd"]     = "/etc/init.d/nginx start",
          ["stop_cmd"]      = "/etc/init.d/nginx stop",
          ["reload_cmd"]    = "/etc/init.d/nginx reload",
          ["restart_cmd"]   = "/etc/init.d/nginx restart",
          ["available_dir"] = "/etc/nginx/sites-available",
          ["enabled_dir"]   = "/etc/nginx/sites-enabled",
          ["enable_cmd"]    = nil,
          ["disable_cmd"]   = nil,
          ["httpd_user"]    = "www-data",
          ["httpd_group"]   = "www-data",
          ["ssl_chain_dir"] = "/etc/ssl/chains",
          ["ssl_cert_dir"]  = "/etc/ssl/certs",
          ["ssl_key_dir"]   = "/etc/ssl/private",
          ["ssl_key_user"]  = "root",
          ["ssl_key_group"] = "ssl-cert",
          ["ssl_key_mode"]  = "0640",
          ["htdocs_path"]   = LC.web.HTDOCS_PATH,
          ["units"]         = { ["nginx"] = "nginx", ["lclogsplit"] = "lclogsplit" }
        }
        check_http2(data)
        return data
      end
      -- else: fall trough, to check for custom NGINX installation
      LC.log.print(LC.log.DEBUG, "LC.nginx.detect(): Found NGINX package '", pkg, "', but no binary at ", bin)
    end
  elseif LC.distribution.family == "Gentoo" then
    -- Debian/Ubuntu
    -- Package is named "nginx"
    local pkg, v = LC.distribution.hasPackage( 'nginx')
    if pkg ~= nil then
      LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
      local bin = "/usr/sbin/nginx"
      -- get binary version
      local bv = getBinaryVersion(bin)
      LC.log.print(LC.log.DEBUG, "binver ", bv)
      if bv ~= nil then
        -- ok, we have all informations. return data table:
        local data = {
          ["type"]          = "nginx",
          ["binname"]       = bin,
          ["binversion"]    = bv,
          ["pkgname"]       = pkg,
          ["pkgversion"]    = v,
          ["configpath"]    = "/etc/nginx",
          ["configfile"]    = "/etc/nginx/conf.d/liveconfig.conf",
          ["statusfile"]    = "/etc/nginx/liveconfig.status",
          ["start_cmd"]     = "/etc/init.d/nginx start",
          ["stop_cmd"]      = "/etc/init.d/nginx stop",
          ["reload_cmd"]    = "/etc/init.d/nginx reload",
          ["restart_cmd"]   = "/etc/init.d/nginx restart",
          ["available_dir"] = "/etc/nginx/sites-available",
          ["enabled_dir"]   = "/etc/nginx/sites-enabled",
          ["enable_cmd"]    = nil,
          ["disable_cmd"]   = nil,
          ["httpd_user"]    = "nginx",
          ["httpd_group"]   = "nginx",
          ["ssl_chain_dir"] = "/etc/ssl/certs",
          ["ssl_cert_dir"]  = "/etc/ssl/certs",
          ["ssl_key_dir"]   = "/etc/ssl/private",
          ["ssl_key_user"]  = "root",
          ["ssl_key_group"] = "root",
          ["ssl_key_mode"]  = "0640",
          ["htdocs_path"]   = LC.web.HTDOCS_PATH,
          ["units"]         = { ["nginx"] = "nginx", ["lclogsplit"] = "lclogsplit" }
        }
        check_http2(data)
        return data
      end
      -- else: fall trough, to check for custom NGINX installation
      LC.log.print(LC.log.DEBUG, "LC.nginx.detect(): Found NGINX package '", pkg, "', but no binary at ", bin)
    end
  elseif LC.distribution.family == "SUSE" then
    -- OpenSUSE
    -- Package is named "nginx"
    local pkg, v = LC.distribution.hasPackage( 'nginx')
    if pkg ~= nil then
      LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
      local bin = "/usr/sbin/nginx"
      -- get binary version
      local bv = getBinaryVersion(bin)
      LC.log.print(LC.log.DEBUG, "binver ", bv)
      if bv ~= nil then
        -- ok, we have all informations. return data table:
        local data = {
          ["type"]          = "nginx",
          ["binname"]       = bin,
          ["binversion"]    = bv,
          ["pkgname"]       = pkg,
          ["pkgversion"]    = v,
          ["configpath"]    = "/etc/nginx",
          ["configfile"]    = "/etc/nginx/conf.d/liveconfig.conf",
          ["statusfile"]    = "/etc/nginx/liveconfig.status",
          ["start_cmd"]     = "/sbin/service nginx start",
          ["stop_cmd"]      = "/sbin/service nginx stop",
          ["reload_cmd"]    = "/sbin/service nginx reload",
          ["restart_cmd"]   = "/sbin/service nginx restart",
--          ["available_dir"] = "/etc/nginx/sites-available",
          ["enabled_dir"]   = "/etc/nginx/vhosts.d",
          ["enable_cmd"]    = nil,
          ["disable_cmd"]   = nil,
          ["httpd_user"]    = "www-data",
          ["httpd_group"]   = "www-data",
          ["ssl_chain_dir"] = "/etc/nginx/ssl",
          ["ssl_cert_dir"]  = "/etc/nginx/ssl",
          ["ssl_key_dir"]   = "/etc/nginx/ssl",
          ["ssl_key_user"]  = "root",
          ["ssl_key_group"] = "root",
          ["ssl_key_mode"]  = "0600",
          ["htdocs_path"]   = LC.web.HTDOCS_PATH,
          ["units"]         = { ["nginx"] = "nginx", ["lclogsplit"] = "lclogsplit" }
        }
        check_http2(data)
        return data
      end
      -- else: fall trough, to check for custom NGINX installation
      LC.log.print(LC.log.DEBUG, "LC.nginx.detect(): Found NGINX package '", pkg, "', but no binary at ", bin)
    end
  elseif LC.distribution.family == "RedHat" then
    -- RedHat/CentOS
    -- Package is named "nginx", available at http://nginx.org/packages/centos/
    local pkg, v = LC.distribution.hasPackage('nginx')
    if pkg ~= nil then
      LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
      local bin = "/usr/sbin/nginx"
      -- get binary version
      local bv = getBinaryVersion(bin)
      LC.log.print(LC.log.DEBUG, "binver ", bv)
      if bv ~= nil then
        -- ok, we have all informations. return data table:
        local data = {
          ["type"]          = "nginx",
          ["binname"]       = bin,
          ["binversion"]    = bv,
          ["pkgname"]       = pkg,
          ["pkgversion"]    = v,
          ["configpath"]    = "/etc/nginx",
          ["configfile"]    = "/etc/nginx/conf.d/liveconfig.conf",
          ["statusfile"]    = "/etc/nginx/liveconfig.status",
          ["start_cmd"]     = "/sbin/service nginx start",
          ["stop_cmd"]      = "/sbin/service nginx stop",
          ["reload_cmd"]    = "/sbin/service nginx reload",
          ["restart_cmd"]   = "/sbin/service nginx restart",
--          ["available_dir"] = "/etc/nginx/sites-available",
          ["enabled_dir"]   = "/etc/nginx/vhosts.d",
          ["enable_cmd"]    = nil,
          ["disable_cmd"]   = nil,
          ["httpd_user"]    = "apache",
          ["httpd_group"]   = "apache",
          ["ssl_chain_dir"] = "/etc/pki/tls/certs",
          ["ssl_cert_dir"]  = "/etc/pki/tls/certs",
          ["ssl_key_dir"]   = "/etc/pki/tls/private",
          ["ssl_key_user"]  = "root",
          ["ssl_key_group"] = "root",
          ["ssl_key_mode"]  = "0600",
          ["htdocs_path"]   = LC.web.HTDOCS_PATH,
          ["units"]         = { ["nginx"] = "nginx", ["lclogsplit"] = "lclogsplit" }
        }
        check_http2(data)
        return data
      end
      -- else: fall trough, to check for custom NGINX installation
      LC.log.print(LC.log.DEBUG, "LC.nginx.detect(): Found NGINX package '", pkg, "', but no binary at ", bin)
    end
  end  
end

-- ---------------------------------------------------------------------------
-- lclogsplit_enable
-- ---------------------------------------------------------------------------
function lclogsplit_enable(cfg)
  -- install service (systemd)
  if LC.fs.is_dir("/lib/systemd/system") and not LC.fs.is_file("/lib/systemd/system/lclogsplit.service") then
    LC.exec("cp --preserve=all " .. LC.liveconfig.libexecdir .. "/lclogsplit.service /lib/systemd/system/")
    if LC.fs.is_file("/bin/systemctl") then
      LC.exec("/bin/systemctl enable lclogsplit")
    end
  end

  -- install init script
  if not LC.fs.is_file("/etc/init.d/lclogsplit") then
    LC.exec("cp " .. LC.liveconfig.libexecdir .. "/lclogsplit.init /etc/init.d/lclogsplit")
    LC.fs.setperm("/etc/init.d/lclogsplit", "0700", "root", "root")

    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d lclogsplit defaults")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --add lclogsplit")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update add lclogsplit default")
    else
      LC.log.print(LC.log.ERR, "Can't install init script 'lclogsplit' - no supported tool found")
    end

  end

  -- start lclogsplit if not running
  if LC.fs.is_file("/sbin/service") then
    LC.exec("/sbin/service lclogsplit start")
  else
    LC.exec("/etc/init.d/lclogsplit start")
  end
end

-- ---------------------------------------------------------------------------
-- lclogsplit_disable
-- ---------------------------------------------------------------------------
function lclogsplit_disable(cfg)

  -- disable lclogsplit (chkconfig)
  if LC.fs.is_file("/etc/init.d/lclogsplit") then

    -- quit lclogsplit
    if LC.fs.is_file("/sbin/service") then
      LC.exec("/sbin/service lclogsplit stop")
    else
      LC.exec("/etc/init.d/lclogsplit stop")
    end

    if LC.fs.is_file("/lib/systemd/system/lclogsplit") then
      -- disable service (systemd)
      if LC.fs.is_file("/bin/systemctl") then
        LC.exec("/bin/systemctl disable lclogsplit")
      end
      -- remove service file
      os.remove("/lib/systemd/system/lclogsplit")
    end

    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d -f lclogsplit remove")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --del lclogsplit")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update del lclogsplit default")
    else
      LC.log.print(LC.log.ERR, "Can't uninstall init script 'lclogsplit' - no supported tool found")
    end
    os.remove("/etc/init.d/lclogsplit")
  end
end

-- ---------------------------------------------------------------------------
-- install(cfg, opts)
--
-- Install LiveConfig configuration files for NGINX
-- ---------------------------------------------------------------------------
function install(cfg, opts)
  LC.log.print(LC.log.INFO, "Activating management of NGINX")

  local fh, status, msg
  local configpath = cfg["configpath"]
  local configfile = cfg["configfile"]
  local statusfile = cfg["statusfile"]
  local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
  local vhostfile  -- set below (distribution-specific)

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    configpath = opts.prefix .. configpath
    configfile = opts.prefix .. configfile
    statusfile = opts.prefix .. statusfile
    vhostpath  = opts.prefix .. vhostpath
  end

  -- create directory for liveconfig.conf:
  if not LC.fs.is_dir(configpath .. "/conf.d") then
    LC.fs.mkdir(configpath .. "/conf.d")
    LC.fs.setperm(configpath .. "/conf.d", 755, "root", "root")
  end

  -- create 'liveconfig.conf' config file
  fh, msg = io.open(configfile, "w")
  if fh == nil then
    LC.log.print(LC.log.ERR, "Can't open '", configfile, "' for writing: ", msg)
    return false, "Can't open '" .. configfile .. "' for writing: " .. msg
  end

  -- write header
  LC.liveconfig.writeHeader(fh)

  -- write config...
  fh:write([[

# Access logging
# We define our own log format here, containing the "canonical" server name

log_format LiveConfig '$server_name:#:$remote_addr - $remote_user [$time_local] '
                      '"$request" $status $body_bytes_sent '
                      '"$http_referer" "$http_user_agent" $request_length $bytes_sent';

access_log /var/log/nginx/access.liveconfig.log LiveConfig;

]])

  if LC.distribution.family == "RedHat" then
    fh:write([[
# Virtual Hosts configuration
include /etc/nginx/vhosts.d/*;

]])
    if LC.fs.is_file("/etc/nginx/conf.d/default.conf") then
      -- move away default.conf, as this contains a "listen" command:
      LC.fs.rename("/etc/nginx/conf.d/default.conf", "/etc/nginx/conf.d/default.conf.lcbak")
    end
  end

  -- write footer
  LC.liveconfig.writeFooter(fh)

  fh:close()

  -- tune/fix original config file
  if LC.fs.is_file(configpath .. "/nginx.conf") then
    -- disable "ssl_prefer_server_ciphers" and "ssl_protocols" in nginx.conf - we're setting this in our own configuration!
    LC.exec("sed -i -e 's/^\\(\\s*\\(ssl_prefer_server_ciphers\\|ssl_protocols\\)\\)/#\\1/' " .. configpath .. "/nginx.conf")
  end

  -- create directory for vhosts:
  if not LC.fs.is_dir(vhostpath) then
    LC.fs.mkdir(vhostpath)
    LC.fs.setperm(vhostpath, 750, "root", cfg.httpd_group)
  end

  if LC.distribution.family == "Debian" then
    vhostfile = "default"
  elseif LC.distribution.family == "Gentoo" then
    vhostfile = "default"
  elseif LC.distribution.family == "SUSE" then
    -- ###TODO### replace nginx.conf (this creates a default server we don't want)
    vhostfile = "default.conf"
  elseif LC.distribution.family == "RedHat" then
    vhostfile = "default.conf"
  else
    LC.log.print(LC.log.ERR, "Distribution family '", LC.distribution.family, "' not supported (yet)")
    return false, "Distribution family '" .. LC.distribution.family .. "' not supported (yet)"
  end

  -- back up old default vhost file (if existing)
  if LC.fs.is_file(vhostpath .. "/" .. vhostfile) and not LC.fs.is_file(vhostpath .. "/" .. vhostfile .. ".lcbak") then
    LC.fs.rename(vhostpath .. "/" .. vhostfile, vhostpath .. "/" .. vhostfile .. ".lcbak")
  end

  -- due to an odd behaviour, we need to *restart* NGINX after initial configuration.
  flag_restart = true

  -- write default vhost file:
  configure(cfg, opts)

  -- install init script for controlling PHP-FCGI instances
  if not LC.fs.is_file("/etc/init.d/nginx-php-fcgi") then
    LC.exec("cp " .. LC.liveconfig.libexecdir .. "/nginx-php-fcgi /etc/init.d/nginx-php-fcgi")

    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d nginx-php-fcgi defaults")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --add nginx-php-fcgi")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update add nginx-php-fcgi default")
    else
      LC.log.print(LC.log.ERR, "Can't install init script 'nginx-php-fcgi' - no supported tool found")
    end
  end

  -- install lclogsplit service:
  lclogsplit_enable()

  -- write status file
  LC.liveconfig.writeStatus(statusfile, cfg.type, 0, os.date("%Y-%m-%d %H:%M:%S %Z"))

  return true
end

-- ---------------------------------------------------------------------------
-- uninstall(cfg, opts)
--
-- Remove LiveConfig configuration files from NGINX
-- ---------------------------------------------------------------------------
function uninstall(cfg, opts)
  LC.log.print(LC.log.INFO, "Disabling management of NGINX")

  local status, msg
  local configpath = cfg["configpath"]
  local configfile = cfg["configfile"]
  local statusfile = cfg["statusfile"]

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    configpath = opts.prefix .. configpath
    configfile = opts.prefix .. configfile
    statusfile = opts.prefix .. statusfile
  end

  if LC.distribution.family == "Debian" then
  elseif LC.distribution.family == "Gentoo" then
  elseif LC.distribution.family == "SUSE" then
  else
    LC.log.print(LC.log.ERR, "Distribution family '", LC.distribution.family, "' not supported (yet)")
    return false, "Distribution family '" .. LC.distribution.family .. "' not supported (yet)"
  end

  -- remove init script for PHP FCGI
  if LC.fs.is_file("/etc/init.d/nginx-php-fcgi") then
    LC.exec("/etc/init.d/nginx-php-fcgi stop")

    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d -f nginx-php-fcgi remove")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --del nginx-php-fcgi")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update del nginx-php-fcgi default")
    else
      LC.log.print(LC.log.ERR, "Can't uninstall init script 'nginx-php-fcgi' - no supported tool found")
    end
    os.remove("/etc/init.d/nginx-php-fcgi")
  end

  -- disable lclogsplit service
  lclogsplit_disable()

  -- remove liveconfig.conf:
  os.remove(configfile)

  -- remove status file
  status, msg = os.remove(statusfile)
  if status == nil then
    return false, LC.log.print(LC.log.ERR, "Deletion of '", statusfile, "' failed: ", msg)
  end

  return true
end


-- ---------------------------------------------------------------------------
-- configure(cfg, opts)
--
-- Configure NGINX
-- ---------------------------------------------------------------------------
function configure(cfg, opts)

  local fh, status, msg
  local configpath = cfg["configpath"]
  local statusfile = cfg["statusfile"]
  local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
  local vhostfile  -- set below (distribution-specific)

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    configpath = opts.prefix .. configpath
    statusfile = opts.prefix .. statusfile
    vhostpath  = opts.prefix .. vhostpath
  end

  local v_major, v_minor, v_patch = string.match(cfg.binversion, "(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)

  if LC.distribution.family == "Debian" then
    vhostfile = "default"
  elseif LC.distribution.family == "Gentoo" or LC.distribution.family == "RedHat" then
    vhostfile = "default"
  elseif LC.distribution.family == "SUSE" then
    vhostfile = "default.conf"
  else
    LC.log.print(LC.log.ERR, "Distribution family '", LC.distribution.family, "' not supported (yet)2")
    return false, "Distribution family '" .. LC.distribution.family .. "' not supported (yet)2"
  end

  fh, msg = io.open(vhostpath .. "/" .. vhostfile, "w")
  if fh == nil then
    LC.log.print(LC.log.ERR, "Can't open '", vhostpath .. "/" .. vhostfile, "' for writing: ", msg)
    return false, "Can't open '" .. vhostpath .. "/" .. vhostfile.. "' for writing: " .. msg
  end
  LC.liveconfig.writeHeader(fh)

  if opts.interfaces ~= nil or opts.ssl_interfaces ~= nil then
    fh:write("# Automatically created by LiveConfig\n\n")

    -- proxy/cache configuration
    fh:write("# default proxy/cache parameters:\n")
    fh:write("proxy_cache_path /var/cache/nginx/cache levels=1:2 keys_zone=cache:64m inactive=60m max_size=10g;\n")
    fh:write("\n")

    if not LC.fs.is_dir("/var/cache/nginx") then
      LC.fs.mkdir("/var/cache/nginx")
      LC.fs.setperm("/var/cache/nginx", 700, cfg["httpd_user"], cfg["httpd_group"])
    end

    if not LC.fs.is_dir("/var/cache/nginx/cache") then
      LC.fs.mkdir("/var/cache/nginx/cache")
      LC.fs.setperm("/var/cache/nginx/cache", 700, cfg["httpd_user"], cfg["httpd_group"])
    end

    if not LC.fs.is_dir("/var/cache/nginx/tmp") then
      LC.fs.mkdir("/var/cache/nginx/tmp")
      LC.fs.setperm("/var/cache/nginx/tmp", 700, cfg["httpd_user"], cfg["httpd_group"])
    end

    if opts.ssl_interfaces ~= nil then
      fh:write("# default SSL parameters:\n")
      fh:write("ssl_prefer_server_ciphers on;\n")
      if opts.ssl_strong then
        fh:write("ssl_ciphers ", LC.liveconfig.DEFAULT_SSL_STRONG_CIPHERS, ";\n")
      else
        fh:write("ssl_ciphers ", LC.liveconfig.DEFAULT_SSL_CIPHERS, ";\n")
      end

      local protocols = ""
      if not opts.ssl_notls1 and not opts.ssl_strong then protocols = " TLSv1" end
      if v_major > 1 or (v_major == 1 and (v_minor > 1 or (v_minor == 1 and v_patch >= 13) or (v_minor == 0 and v_patch >= 12))) then
        -- "TLSv1.1" and "TLSv1.2" are supported from 1.1.13+ and 1.0.12+, see http://nginx.org/en/docs/http/ngx_http_ssl_module.html#ssl_protocols
        if not opts.ssl_notls1 and not opts.ssl_strong then protocols = protocols .. " TLSv1.1" end
        protocols = protocols .. " TLSv1.2"
        if v_major > 1 or (v_major == 1 and v_minor >= 13 ) then
          protocols = protocols .. " TLSv1.3"
        end
      end

      if protocols ~= "" then
        fh:write("ssl_protocols", protocols, ";\n")
      else
        fh:write("# WARNING: no valid configuration of 'ssl_protocols' possible... using default configuration.\n")
      end

      -- enable SSL session cache
      fh:write("ssl_session_cache shared:ssl_session_cache:10m;\n")

      -- custom DH parameters (Logjam attack) - requires v0.7.2+
      if v_major > 0 or (v_major == 0 and v_minor > 7) or (v_major == 0 and v_minor == 7 and v_patch >= 2) then
        -- do custom DH parameters already exist?
        if not LC.fs.is_file(configpath .. "/dhparam.pem") and LC.fs.is_file("/etc/liveconfig/dhparam.dist.pem") then
          -- copy non-standard default DH parameters until local parameters have been generated:
          LC.fs.copy("/etc/liveconfig/dhparam.dist.pem", configpath .. "/dhparam.pem")
          LC.fs.setperm(configpath .. "/dhparam.pem", "0640", "root", "root")
        end
        if LC.fs.is_file(configpath .. "/dhparam.pem") then
          fh:write("ssl_dhparam ", configpath, "/dhparam.pem;\n")
        end
      end

      -- configure default SSL certificate for shared SSL interfaces (SNI)
      if opts.ssl_sni ~= nil and opts.ssl_sni.interfaces ~= nil then
        -- create/update SSL certificate file:
        local sfh
        local fname = cfg.ssl_cert_dir .. "/httpd-default-combined.crt"
        sfh, msg = io.open(fname, "w")
        if sfh == nil then
          LC.log.print(LC.log.ERR, "Can't open '", fname, "' for writing: ", msg)
          return false, "Can't open '" .. fname .. "' for writing: " .. msg
        end
        sfh:write(opts.ssl_sni.cert)
        if opts.ssl_sni.cachain ~= nil then
          -- append intermediate certificate(s)
          sfh:write(opts.ssl_sni.cachain)
        end
        sfh:close()
        LC.fs.setperm(fname, "0644", "root", "root")

        -- create/update SSL key file:
        fname = cfg.ssl_key_dir .. "/httpd-default.key"
        sfh, msg = io.open(fname, "w")
        if sfh == nil then
          LC.log.print(LC.log.ERR, "Can't open '", fname, "' for writing: ", msg)
          return false, "Can't open '" .. fname .. "' for writing: " .. msg
        end
        sfh:write(opts.ssl_sni.key)
        sfh:close()
        LC.fs.setperm(fname, cfg["ssl_key_mode"], cfg["ssl_key_user"], cfg["ssl_key_group"])

        fh:write("\n")

        -- configure shared SSL interfaces (SNI) to use default vHost with default SSL certificate
        fh:write("# SSL interfaces and default HTTPS server (SNI):\n")
        fh:write("server {\n")
        for i=1, #opts.ssl_sni.interfaces do
          fh:write("\tlisten\t\t", opts.ssl_sni.interfaces[i])
          if v_major == 0 and (v_minor < 8 or (v_minor == 8 and v_patch < 21)) then
            fh:write(" default")
          else
            fh:write(" default_server")
          end
          fh:write(" ssl;\n")
        end

        fh:write("\tssl_certificate ", cfg.ssl_cert_dir, "/httpd-default-combined.crt;\n")
        fh:write("\tssl_certificate_key ", cfg.ssl_key_dir, "/httpd-default.key;\n")

        fh:write([[
  server_name	_;
  root		/usr/share/liveconfig/html;
  ssi		on;
  error_page	403 /not-available-ssl.shtml;
  error_page	404 /not-available-ssl.shtml;
]])

        if v_major < 1 or (v_major == 1 and v_minor < 5) or (v_major == 1 and v_minor == 5 and v_patch < 7) then
          -- mitigate CVE-2013-4547:
          fh:write("\tif ($request_uri ~ \" \") {\n")
          fh:write("\t\treturn 444;\n")
          fh:write("\t}\n")
        end

        fh:write("}\n")

      end

      fh:write("\n")

    end

    -- default server configuration:
    fh:write("# interfaces and default server configuration:\n")
    fh:write("server {\n")

    -- HTTP interfaces:
    if opts.interfaces ~= nil then
      for i=1, #opts.interfaces do
        fh:write("\tlisten\t\t", opts.interfaces[i])
        if v_major == 0 and (v_minor < 8 or (v_minor == 8 and v_patch < 21)) then
          fh:write(" default;\n")
        else
          fh:write(" default_server;\n")
        end
      end
    end

    fh:write([[
	server_name	_;
	root		/usr/share/liveconfig/html;
	ssi		on;
	error_page	403 /not-available.shtml;
	error_page	404 /not-available.shtml;

	location /.well-known/ {
		try_files $uri $uri/ @wellknown;
	}
	location @wellknown {
		rewrite ^/.well-known/(.*/)?([^/]*)$ /$1$host.$2 break;
		root ]], LC.web.getWebRoot(), [[/.well-known/htdocs;
		default_type text/plain;
		auth_basic off;
	}

]])

    if v_major < 1 or (v_major == 1 and v_minor < 5) or (v_major == 1 and v_minor == 5 and v_patch < 7) then
      -- mitigate CVE-2013-4547:
      fh:write("\tif ($request_uri ~ \" \") {\n")
      fh:write("\t\treturn 444;\n")
      fh:write("\t}\n")
    end

    fh:write("\n}\n")
  end

  LC.liveconfig.writeFooter(fh)
  fh:close()

  -- update resolver configuration (/etc/nginx/conf.d/resolver.conf)
  updateResolver()

  -- update status file
  LC.liveconfig.writeStatus(statusfile, cfg.type, opts.revision, os.date("%Y-%m-%d %H:%M:%S %Z"))

  -- reload web server configuration
  LC.timeout.set('nginx.reload', 10, 60)

  return true
end

-- ---------------------------------------------------------------------------
-- Helper function:
-- ---------------------------------------------------------------------------
local function write_server(cfg, opts, fh, vhost, dom, cnames)
  local i, key
  local v_major, v_minor, v_patch = string.match(cfg.binversion, "(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)

  fh:write("server {\n")
  for i, key in ipairs(vhost.ips) do
    fh:write("\tlisten\t\t")
    if vhost.ssl then
      if string.find(key, ":") then
        -- IPv6 address
        fh:write("[", key, "]:443")
      else
        -- IPv4 address
        fh:write(key, ":443")
      end
    else
      if string.find(key, ":") then
        -- IPv6 address
        fh:write("[", key, "]:80")
      else
        -- IPv4 address
        fh:write(key, ":80")
      end
    end
    if vhost.ssl then
      fh:write(" ssl")
      if cfg.has_http2 and vhost.http2 then
        -- enable HTTP/2
        fh:write(" http2")
      end
    end
    fh:write(";\n")
  end

  table.sort(dom)

  for i, key in ipairs(dom) do
    if i == 1 then
      fh:write("\tserver_name\t")
    else
      fh:write("\n\t\t\t")
    end
    fh:write(key)
    cnames[#cnames+1] = key
  end
  fh:write(";\n")

  if vhost.ssl then
    -- SSL configuration:

    local protocols = ""
    if not vhost.ssl_notls1 and not vhost.ssl_strong then protocols = " TLSv1" end
    if v_major > 1 or (v_major == 1 and (v_minor > 1 or (v_minor == 1 and v_patch >= 13) or (v_minor == 0 and v_patch >= 12))) then
      -- "TLSv1.1" and "TLSv1.2" are supported from 1.1.13+ and 1.0.12+, see http://nginx.org/en/docs/http/ngx_http_ssl_module.html#ssl_protocols
      if not vhost.ssl_notls1 and not vhost.ssl_strong then protocols = protocols .. " TLSv1.1" end
      protocols = protocols .. " TLSv1.2"
      if v_major > 1 or (v_major == 1 and v_minor >= 13 ) then
        protocols = protocols .. " TLSv1.3"
      end
    end

    if protocols ~= "" then
      fh:write("\tssl_protocols", protocols, ";\n")
    else
      fh:write("\t# WARNING: no valid configuration of 'ssl_protocols' possible... using default configuration.\n")
    end

    -- create/update SSL certificate file:
    if not LC.fs.is_dir(cfg.ssl_cert_dir) then
      LC.fs.mkdir(cfg.ssl_cert_dir)
      LC.fs.setperm(cfg.ssl_cert_dir, 755, "root", "root")
    end
    local sfh
    local fname = cfg.ssl_cert_dir .. "/" .. vhost.ssl_filename .. "-combined.crt"
    sfh, msg = io.open(fname, "w")
    if sfh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", fname, "' for writing: ", msg)
      return false, "Can't open '" .. fname .. "' for writing: " .. msg
    end
    sfh:write(vhost.ssl_crt)
    sfh:write("\n")
    if vhost.ssl_ca ~= nil then
      -- append intermediate certificates
      sfh:write(vhost.ssl_ca)
      sfh:write("\n")
    end
    sfh:close()
    LC.fs.setperm(fname, "0644", "root", "root")
    fh:write("\tssl_certificate\t", fname, ";\n")

    -- create/update SSL key file:
    if not LC.fs.is_dir(cfg.ssl_key_dir) then
      LC.fs.mkdir(cfg.ssl_key_dir)
      LC.fs.setperm(cfg.ssl_key_dir, 755, "root", "root")
    end
    fname = cfg.ssl_key_dir .. "/" .. vhost.ssl_filename .. ".key"
    sfh, msg = io.open(fname, "w")
    if sfh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", fname, "' for writing: ", msg)
      return false, "Can't open '" .. fname .. "' for writing: " .. msg
    end
    sfh:write(vhost.ssl_key)
    sfh:close()
    LC.fs.setperm(fname, cfg["ssl_key_mode"], cfg["ssl_key_user"], cfg["ssl_key_group"])
    fh:write("\tssl_certificate_key\t", fname, ";\n")

    -- configure OCSP (if supported by NGINX and by the certificate)
    if vhost.ssl_ocsp and (v_major > 1 or (v_major == 1 and (v_minor > 3 or (v_minor == 3 and v_patch >= 7)))) then
      fh:write("\tssl_stapling on;\n")
    end

    fh:write("\n")
  end

  -- /.well-known/ hook for domain validation (SSL certificates):
  fh:write("\tlocation /.well-known/ {\n")
  fh:write("\t\ttry_files $uri $uri/ @wellknown;\n")
  fh:write("\t}\n")
  fh:write("\tlocation @wellknown {\n")
  fh:write("\t\trewrite ^/.well-known/(.*/)?([^/]*)$ /$1$host.$2 break;\n")
  fh:write("\t\troot ", LC.web.getWebRoot(), "/.well-known/htdocs;\n")
  fh:write("\t\tdefault_type text/plain;\n")
  fh:write("\t\tauth_basic off;\n")
  fh:write("\t}\n")

  -- write ACME challenges
  -- only for backward-compatibility with "old" Let's Encrypt / ACME interface!
  local acmeCount = 0
  for i, key in ipairs(dom) do
    if vhost.domains[key].acme ~= nil then
      local acme = vhost.domains[key].acme
      local idx
      for idx = 1, #acme do
        LC.web.addValidation(key, "acme-challenge/" .. acme[idx].token, acme[idx].response)
      end
    end
  end

  -- Web statistics:
  if opts.webstats and opts.webstats.domain and opts.webstats.path and LC.liveconfig.inTable(opts.webstats.domain, dom) then
    fh:write("\t# Web statistics:\n")
    fh:write("\tlocation ^~ /", opts.webstats.path, "/ {\n")
    fh:write("\t\tauth_basic \"Web Statistics\";\n")
    fh:write("\t\tauth_basic_user_file \"", opts.path, "/stats/.htpasswd\";\n")
    fh:write("\t\talias \"", opts.path, "/stats/\";\n")
    fh:write("\t}\n")
    if opts.webstats.software == 1 then
      -- shared resources for Webalizer
      fh:write("\tlocation ^~ /", opts.webstats.path, "/images/ {\n")
      fh:write("\t\talias /usr/share/liveconfig/html/webalizer/;\n")
      fh:write("\t}\n")
    elseif opts.webstats.software == 2 then
      -- shared resources for AWStats
      local icondir = "/usr/share/awstats/icon/"
      if LC.distribution.family == "RedHat" then
        icondir = "/var/www/awstats/icon/"
      elseif LC.distribution.family == "SUSE" then
        icondir = "/usr/share/awstats/icon/"
      elseif LC.distribution.family == "Gentoo" then
        icondir = "/usr/share/awstats/wwwroot/icon/"
      end
      fh:write("\tlocation ^~ /", opts.webstats.path, "/icon/ {\n")
      fh:write("\t\talias ", icondir, ";\n")
      fh:write("\t}\n")
      fh:write("\tlocation ^~ /", opts.webstats.path, "/awstats/ {\n")
      fh:write("\t\talias /usr/share/liveconfig/html/awstats/;\n")
      fh:write("\t}\n")
    end
    fh:write("\n")
  end

end

-- ---------------------------------------------------------------------------
-- fh: file handle
-- opts: configuration options (from LiveConfig)
-- php: php version code (eg. "", "php56", ...)
local function writePHPconfig(fh, opts, php, phpVersionsFCGI)

  local suffix = "." .. php
  if php == "" or php == LC.web.getPHPDefaultCode() then suffix = "" end

  fh:write("\t# PHP configuration for this subscription: ")
  if opts.hasPHP == 2 then
    -- FastCGI (DEPRECATED!)
    fh:write("FastCGI\n")
    fh:write("\tindex index.php index.html index.htm;\n")
    fh:write("\tlocation ~ \\.php(/|$) {\n")
    fh:write("\t\ttry_files $fastcgi_script_name =404;\n")
    fh:write("\t\tfastcgi_split_path_info ^(.+\\.php)(/.*)$;\n")
    fh:write("\t\tinclude /etc/nginx/fastcgi_params;\n")
    fh:write("\t\tset $path_info $fastcgi_path_info;\n")
    fh:write("\t\tfastcgi_param PATH_INFO $path_info;\n")
    fh:write("\t\tfastcgi_param SCRIPT_FILENAME $document_root$fastcgi_script_name;\n")
    fh:write("\t\tfastcgi_index index.php;\n")
    fh:write("\t\tfastcgi_pass unix:", opts.path, "/conf/sockets/nginx-php-fcgi", suffix, ".sock;\n")
    fh:write("\t}\n")
    phpVersionsFCGI[php] = true
  elseif opts.hasPHP == 4 then
    -- FPM
    fh:write("FPM\n")
    fh:write("\tindex index.php index.html index.htm;\n")
    fh:write("\tlocation ~ \\.php(/|$) {\n")

    local p, found
    for p in pairs(LCS.web.phpVersions) do
      if LCS.web.phpVersions[p].code == php or (php == '' and LCS.web.phpVersions[p].code == LC.web.getPHPDefaultCode()) then
        local phpconfig = LCS.web.phpVersions[p]
        if phpconfig.fpm == nil then break end
        found = true
        fh:write("\t\ttry_files $fastcgi_script_name =404;\n")
        fh:write("\t\tfastcgi_split_path_info ^(.+\\.php)(/.*)$;\n")
        fh:write("\t\tinclude /etc/nginx/fastcgi_params;\n")
        fh:write("\t\tset $path_info $fastcgi_path_info;\n")
        fh:write("\t\tfastcgi_param PATH_INFO $path_info;\n")
        fh:write("\t\tfastcgi_param SCRIPT_FILENAME $document_root$fastcgi_script_name;\n")
        fh:write("\t\tfastcgi_index index.php;\n")
        -- socket location must be completely lowercase for Apache compatibility!
        fh:write("\t\tfastcgi_pass unix:", phpconfig.fpm.sockets, "/", opts.name:lower(), ".sock;\n")
        break
      end
    end
    if not found then
      LC.log.print(LC.log.ERR, "NGINX: PHP version ", pv, " configured as FPM for subscription ", opts.name, " but no PHP found for this version!")
      -- to prevent raw .php files being served, return error 500 (internal server error):
      fh:write("\t\terror_page 500 /.errorFiles/no-php-fpm.html;\n")
      fh:write("\t\treturn 500;\n")
    end

    fh:write("\t}\n")
  else
    -- unsupported PHP mode!
    fh:write("UNSUPPORTED MODE (", opts.hasPHP, ")\n")
    fh:write("\tlocation ~ \\.php(/|$) {\n")
    fh:write("\t\treturn 500 \"Unsupported PHP mode configured for this domain. Please contact your administrator.\";\n")
    fh:write("\t}\n")
  end

end

-- ---------------------------------------------------------------------------
-- configureVHost(cfg, opts)
--
-- Configure virtual host for NGINX
-- ---------------------------------------------------------------------------
function configureVHost(cfg, opts, cnames)

  LC.log.print(LC.log.DEBUG, "configureVHost(nginx) called")

  local v_major, v_minor, v_patch = string.match(cfg.binversion, "(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)

  local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
  if opts and opts.prefix then vhostpath = opts.prefix .. vhostpath end -- use prefix (for testing etc.)
  local filename = vhostpath .. "/" .. opts["name"] .. ".conf"
  local tempfile = filename .. ".tmp"
--  tempfile = "/tmp/test.conf"
  local i = 0
  local key
  local phpVersionsFCGI = {}  -- list with all PHP versions actually used with FastCGI (not FPM!)
  local has_php = false   -- is "true" if PHP is enabled for at least one domain/vHost

  -- create temporary config file
  local fh, msg = io.open(tempfile, "w")
  if fh == nil then
    LC.log.print(LC.log.ERR, "Can't open '", tempfile, "' for writing: ", msg)
    return false, "Can't open '" .. tempfile .. "' for writing: " .. msg
  end
  LC.liveconfig.writeHeader(fh)
  fh:write("# Created at: ", os.date("%Y-%m-%d %H:%M:%S %Z"), "\n")
  fh:write("# ----------------------------------------------------------------------------\n\n")

  -- now check every single vhost, configure it if it's for this server:
  local vhostIdx
  local vhostCount = 0

  if type(opts.vhosts) == "table" then
    for vhostIdx=1, #opts.vhosts do
      while opts.vhosts[vhostIdx].server == cfg.type do
        local vhost = opts.vhosts[vhostIdx]

        fh:write("# IP group: '", vhost.ipgroup, "'\n")
        if vhost.ips == nil then
          fh:write("# -> no IPs configured!\n")
          break
        end

        -- found a vHost entry to be configured
        vhostCount = vhostCount + 1

        if opts.suspended then
          -- webspace suspended:
          local dn = { }    -- list with all domain/subdomain names
          for key in pairs(vhost.domains) do
            dn[#dn+1] = key
          end
          write_server(cfg, opts, fh, vhost, dn, cnames)
          fh:write("\tlocation /.errorFiles/ {\n")
          fh:write("\t\talias /usr/share/liveconfig/html/;\n")
          fh:write("\t}\n")
          fh:write("\terror_page 503 /.errorFiles/suspended.html;\n")
          fh:write("\tlocation / {\n")
          fh:write("\t\treturn 503;\n")
          fh:write("\t}\n")
          fh:write("}\n\n")
          break
        end

        -- copy all domain names in a table, and then sort them:
        local sd   = { }    -- list with all domains/subdomains and their destination directory
        local sd_count = 0  -- Lua doesn't support counting table entries with non-integer keys :(
        local r301 = { }    -- list with all domains/subdomains with 301-redirect
        local r301_count = 0
        local r302 = { }    -- list with all domains/subdomains with 302-redirect
        local r302_count = 0
        local prx  = { }    -- list with all domains/subdomains with proxy destination
        local prx_count = 0
        local app  = { }    -- list with all domains/subdomains with web application
        local app_count = 0
        local bldr = { }    -- list with all domains/subdomains with website builder destination
        local bldr_count = 0
        local rev  = { }    -- list with all domains/subdomains with reverse proxy destination
        local rev_count = 0
        for key in pairs(vhost.domains) do
          local dst = vhost.domains[key].dest
          if dst == nil then dst = '' end
          local php = vhost.domains[key].php
          if php == nil then php = '' end

          if vhost.domains[key].revproxy then
            -- passed to Apache as reverse proxy
            if rev[dst] == nil then rev[dst] = { } end
            rev[dst][#rev[dst] + 1] = key
            rev_count = rev_count + 1
          elseif vhost.domains[key].type == 1 then
            -- normal (local) (sub)domain; add to subdomain list, grouped by PHP version code
            if sd[php] == nil then sd[php] = { } end
            if sd[php][dst] == nil then sd[php][dst] = { } end
            sd[php][dst][#sd[php][dst] + 1] = key
            sd_count = sd_count + 1
          elseif vhost.domains[key].type == 2 then
            -- (sub)domain with 301-redirect
            if r301[dst] == nil then r301[dst] = { } end
            r301[dst][#r301[dst] + 1] = key
            r301_count = r301_count + 1
          elseif vhost.domains[key].type == 3 then
            -- (sub)domain with 302-redirect
            if r302[dst] == nil then r302[dst] = { } end
            r302[dst][#r302[dst] + 1] = key
            r302_count = r302_count + 1
          elseif vhost.domains[key].type == 4 then
            -- (sub)domain with proxy destination
            if prx[dst] == nil then prx[dst] = { } end
            prx[dst][#prx[dst] + 1] = key
            prx_count = prx_count + 1
          elseif vhost.domains[key].type == 5 then
            -- (sub)domain with web application, grouped by PHP version code!
            if app[php] == nil then app[php] = { } end
            if app[php][dst] == nil then app[php][dst] = { } end
            app[php][dst][#app[php][dst] + 1] = key
            app_count = app_count + 1
          elseif vhost.domains[key].type == 7 then
            -- Website builder
            if bldr[dst] == nil then bldr[dst] = { } end
            bldr[dst][#bldr[dst] + 1] = key
            bldr_count = bldr_count + 1
          end

        end

        -- create server{} sections for all "normal" (sub)domains
        if sd_count > 0 then
          for php in pairs(sd) do
            for key in pairs(sd[php]) do

              table.sort(sd[php][key])

              write_server(cfg, opts, fh, vhost, sd[php][key], cnames)
              fh:write("\troot\t\t\"", opts.path, "/" .. LC.web.HTDOCS_PATH .. "/", escape(key), "\";\n")

              if (v_major > 1 or (v_major == 1 and (v_minor > 1 or (v_minor == 1 and v_patch >= 5)))) then
                fh:write("\tdisable_symlinks if_not_owner;\n")
              end

              -- error.log enabled?
              if opts.errorlog then
                fh:write("\terror_log\t", opts.path, "/logs/nginx.error.log warn;\n\n")
              end

              if opts.hasSSI then
                fh:write("\tssi\t\ton;\n\n")
              end

              if opts.phpini and opts.phpini['post_max_size'] then
                local val = LC.web.php2int(opts.phpini['post_max_size'].value)
                fh:write("\tclient_max_body_size\t", val, ";\n\n")
              end

              -- include custom configuration options
              if LC.fs.is_file(opts.path .. "/conf/nginx.conf") then
                fh:write("\t# Include customer-specific configuration options:\n")
                fh:write("\tinclude ", opts.path, "/conf/nginx.conf;\n\n")
              end

              for i in pairs(sd[php][key]) do
                if LC.fs.is_file(opts.path .. "/conf/" .. sd[php][key][i] .. ".nginx.conf") then
                  fh:write("\t# Include domain-specific configuration options:\n")
                  fh:write("\tinclude ", opts.path, "/conf/", sd[php][key][i], ".nginx.conf;\n\n")
                  break
                end
              end

              -- prevent access to .htacces/.htpasswd files
              fh:write("\tlocation ~ /\\.ht {\n")
              fh:write("\t\tdeny all;\n")
              fh:write("\t}\n\n")

              -- CGI not supported...
              fh:write("\tlocation ~ ^/cgi-bin/ {\n")
              fh:write("\t\tdeny all;\n")
              fh:write("\t}\n\n")

              -- password-protected directories
              -- location is specified with regular expression (not prefix!) to have highest precedence
              -- (see http://nginx.org/en/docs/http/ngx_http_core_module.html#location)
              if opts.pwprotect then
                fh:write("\n\t# password-protected directories:\n")
                local pwpath
                local docroot = normalizeSlashes( "/" .. LC.web.HTDOCS_PATH .. "/" .. escape(key) .. "/" )
                for pwpath in pairs(opts.pwprotect) do
                  local p = normalizeSlashes(pwpath .. "/")
                  if string.sub(p, 1, string.len(docroot)) == docroot then
                    local loc = string.sub(p, string.len(docroot))
                    fh:write("\tlocation ~ ^", loc, " {\n")
                    fh:write("\t\tauth_basic \"", escape(opts.pwprotect[pwpath].title), "\";\n")
                    fh:write("\t\tauth_basic_user_file ", opts.path, "/conf/.htpasswd;\n")
                    if opts.pwprotect[pwpath].users[1] ~= "*" then
                      -- limit access to specific user(s)
                      fh:write("\t\tif ($remote_user !~ ^(")
                      for i in pairs(opts.pwprotect[pwpath].users) do
                        if i > 1 then fh:write("|") end
                        fh:write(escapePCRE(opts.pwprotect[pwpath].users[i]))
                      end
                      fh:write(")?$) {\n")
                      fh:write("\t\t\treturn 403;\n")
                      fh:write("\t\t}\n")
                    end

                    -- special case for password-protected root directory
                    if loc == '/' or loc:sub(1, 13) == '/.well-known/' then
                      fh:write("\t\tlocation ~ ^/.well-known/(.*/)?([^/]*)$ {\n")
                      fh:write("\t\t\terror_page 470 = @wellknown;\n")
                      fh:write("\t\t\tif (-f ", LC.web.getWebRoot(), "/.well-known/htdocs/$1$host.$2) { return 470; }\n")
                      fh:write("\t\t}\n")
                    end

                    -- include PHP settings (otherwise PHP won't be available within password-protected directories)
                    if opts.hasPHP and opts.hasPHP > 0 then
                      -- PHP enabled:
                      writePHPconfig(fh, opts, php, phpVersionsFCGI)
                    end

                    fh:write("\t}\n")
                  end
                end
                fh:write("\n")
              end

              -- PHP settings
              if opts.hasPHP and opts.hasPHP > 0 then
                -- PHP enabled:
                writePHPconfig(fh, opts, php, phpVersionsFCGI)
                has_php = true
              end

              -- add nice default error pages
              if COMING_SOON == 1 or (COMING_SOON == 2 and (key == '' or key == '/')) then
                fh:write("\n\tlocation = / {\n")
                fh:write("\t\terror_page\t403 /.errorFiles/coming-soon.html;\n")
                fh:write("\t}\n")
                fh:write("\tlocation /.errorFiles/ {\n")
                fh:write("\t\talias /usr/share/liveconfig/html/;\n")
                fh:write("\t}\n")
              end

              fh:write("}\n\n")
            end
          end -- for php in pairs(sd) do ...
        end -- if sd_count > 0

        -- create server{} section for all domains with 301-redirects
        if r301_count > 0 then
          for key in pairs(r301) do
            write_server(cfg, opts, fh, vhost, r301[key], cnames)
            fh:write("\tlocation /{\n")
            if string.sub(key, 1, 9) == 'https://*' then
              key = 'https://$host/*'
            else
              key = string.gsub(key, "\\", "\\\\")
              key = string.gsub(key, "\"", "\\\"")
              key = string.gsub(key, "%$", "\\$")
            end
            if string.sub(key, -2) == '/*' then
              fh:write("\t\trewrite\t\t^/(.*)$ \"", string.sub(key, 1, string.len(key)-2), "/$1\" permanent;\n")
            else
              fh:write("\t\trewrite\t\t^/.* \"", key, "\" permanent;\n")
            end
            fh:write("\t}\n")
            fh:write("}\n\n")
          end
        end

        -- create server{} section for all domains with 302-redirects
        if r302_count > 0 then
          for key in pairs(r302) do
            write_server(cfg, opts, fh, vhost, r302[key], cnames)
            fh:write("\tlocation /{\n")
            if string.sub(key, 1, 9) == 'https://*' then
              key = 'https://$host/*'
            else
              key = string.gsub(key, "\\", "\\\\")
              key = string.gsub(key, "\"", "\\\"")
              key = string.gsub(key, "%$", "\\$")
            end
            if string.sub(key, -2) == '/*' then
              fh:write("\t\trewrite\t\t^/(.*)$ \"", string.sub(key, 1, string.len(key)-2), "/$1\" redirect;\n")
            else
              fh:write("\t\trewrite\t\t^/.* \"", key, "\" redirect;\n")
            end
            fh:write("\t}\n")
            fh:write("}\n\n")
          end
        end

        -- create server{} section for all domains with proxy commands
        if prx_count > 0 then

          -- disable HTTP/2 for reverse proxy domains
          vhost.http2 = false

          for key in pairs(prx) do
            write_server(cfg, opts, fh, vhost, prx[key], cnames)
            local proto, host, path = key:match('^([^:]*)://([^/]+)(/?.*)$')
            if proto == nil or host == nil then
              LC.log.print(LC.log.ERR, "nginx.lua: can't parse proxy redirect URL '" .. key .. "'")
            else
              if path:sub(-2) == '/*' then path = path:sub(1, path:len()-2) end
              while path:sub(-1) == '/' do
                path = path:sub(1, path:len()-1)
              end
              fh:write("\tlocation / {\n")
              fh:write("\t\tset $proxy_host ", host, ";\n")
              fh:write("\t\tproxy_pass\t", proto, "://$proxy_host", path, "$request_uri;\n")
              local i, s
              for i,s in pairs(PROXY_PARAMS) do
                if type(s) == 'table' then
                  local j, t
                  for j,t in pairs(s) do fh:write("\t\t", i, "\t", t, ";\n") end
                else
                  fh:write("\t\t", i, "\t", s, ";\n")
                end
              end
              fh:write("\t}\n")
            end
            fh:write("}\n\n")
          end
        end

        if app_count > 0 then
          for php in pairs(app) do
            for key in pairs(app[php]) do
              write_server(cfg, opts, fh, vhost, app[php][key], cnames)
              fh:write("\troot\t\t\"", opts.path, "/apps/", key, "\";\n")

              if (v_major > 1 or (v_major == 1 and (v_minor > 1 or (v_minor == 1 and v_patch >= 5)))) then
                fh:write("\tdisable_symlinks if_not_owner;\n")
              end

              if opts.hasSSI then
                fh:write("\tssi\t\ton;\n\n")
              end

              -- prevent access to .htacces/.htpasswd files
              fh:write("\tlocation ~ /\\.ht {\n")
              fh:write("\t\tdeny all;\n")
              fh:write("\t}\n\n")

              -- CGI not supported...
              fh:write("\tlocation ~ ^/cgi-bin/ {\n")
              fh:write("\t\tdeny all;\n")
              fh:write("\t}\n\n")

              -- PHP settings
              if opts.hasPHP and opts.hasPHP > 0 then
                -- PHP enabled:
                writePHPconfig(fh, opts, php, phpVersionsFCGI)
                has_php = true
              end

              -- include custom configuration options
              if LC.fs.is_file(opts.path .. "/conf/nginx.conf") then
                fh:write("    # Include customer-specific configuration options:\n")
                fh:write("    include ", opts.path, "/conf/nginx.conf;\n\n")
              end

              for i in pairs(app[php][key]) do
                if LC.fs.is_file(opts.path .. "/conf/" .. app[php][key][i] .. ".nginx.conf") then
                  fh:write("\t# Include domain-specific configuration options:\n")
                  fh:write("\tinclude ", opts.path, "/conf/", app[php][key][i], ".nginx.conf;\n\n")
                  break
                end
              end

              fh:write("}\n\n")
            end
          end -- for php in pairs(app) do
        end -- if app_count > 0 then

        if bldr_count > 0 then
          for key in pairs(bldr) do
            write_server(cfg, opts, fh, vhost, bldr[key], cnames)
            fh:write("\troot\t\t\"", opts.path, "/", key, "\";\n")

            if (v_major > 1 or (v_major == 1 and (v_minor > 1 or (v_minor == 1 and v_patch >= 5)))) then
              fh:write("\tdisable_symlinks if_not_owner;\n")
            end

            if opts.hasSSI then
              fh:write("\tssi\t\ton;\n\n")
            end

            -- prevent access to .htacces/.htpasswd files
            fh:write("\tlocation ~ /\\.ht {\n")
            fh:write("\t\tdeny all;\n")
            fh:write("\t}\n\n")

            -- CGI not supported...
            fh:write("\tlocation ~ ^/cgi-bin/ {\n")
            fh:write("\t\tdeny all;\n")
            fh:write("\t}\n\n")

            -- PHP settings
            if opts.hasPHP and opts.hasPHP > 0 then
              -- PHP enabled:
              writePHPconfig(fh, opts, php, phpVersionsFCGI)
              has_php = true
            end

            -- include custom configuration options
            if LC.fs.is_file(opts.path .. "/conf/nginx.conf") then
              fh:write("    # Include customer-specific configuration options:\n")
              fh:write("    include ", opts.path, "/conf/nginx.conf;\n\n")
            end

            for i in pairs(bldr[key]) do
              if LC.fs.is_file(opts.path .. "/conf/" .. bldr[key][i] .. ".nginx.conf") then
                fh:write("\t# Include domain-specific configuration options:\n")
                fh:write("\tinclude ", opts.path, "/conf/", bldr[key][i], ".nginx.conf;\n\n")
                break
              end
            end

            fh:write("}\n\n")
          end
        end -- if bldr_count > 0 then

        -- create server{} section for all reverse proxy domains
        if rev_count > 0 then
          for key in pairs(rev) do
            write_server(cfg, opts, fh, vhost, rev[key], cnames)
            fh:write("\n\t# get all content from Apache\n")
            fh:write("\tlocation / {\n")
            fh:write("\t\tproxy_pass\thttp://127.0.0.1;\n")

            local i, s
            for i,s in pairs(PROXY_PARAMS) do
              if type(s) == 'table' then
                local j, t
                for j,t in pairs(s) do fh:write("\t\t", i, "\t", t, ";\n") end
              else
                fh:write("\t\t", i, "\t", s, ";\n")
              end
            end

            fh:write("\t}\n")
            fh:write("\n\t# cache static content\n")
            fh:write("\tlocation ~* ^.+\\.(jpg|jpeg|gif|png|bmp|ico|zip|tgz|gz|rar|bz2|doc|xls|exe|pdf|ppt|txt|tar|mid|midi|wav|mp3|rtf)$ {\n")
            fh:write("\t\tproxy_pass\thttp://127.0.0.1;\n")

            local i, s
            for i,s in pairs(PROXY_PARAMS) do
              if type(s) == 'table' then
                local j, t
                for j,t in pairs(s) do fh:write("\t\t", i, "\t", t, ";\n") end
              else
                fh:write("\t\t", i, "\t", s, ";\n")
              end
            end

            fh:write("\t\tproxy_cache\tcache;\n")
            fh:write("\t\tproxy_cache_key\t\"$scheme://$host$request_uri\";\n")
            fh:write("\t\tproxy_cache_min_uses\t2;\n")
            fh:write("\t\tproxy_cache_valid\t1800s;\n")
            fh:write("\t\tproxy_cache_valid\t404\t10s;\n")
            fh:write("\t\tproxy_cache_use_stale\terror timeout invalid_header updating http_500 http_502 http_503 http_504;\n")
            fh:write("\t}\n")

            -- password-protected directories
            -- location is specified with regular expression (not prefix!) to have highest precedence
            -- (see http://nginx.org/en/docs/http/ngx_http_core_module.html#location)
            if opts.pwprotect then
              fh:write("\n\t# password-protected directories:\n")
              local pwpath
              local docroot = normalizeSlashes( "/" .. LC.web.HTDOCS_PATH .. "/" .. escape(key) .. "/" )
              for pwpath in pairs(opts.pwprotect) do
                local p = normalizeSlashes(pwpath .. "/")
                if string.sub(p, 1, string.len(docroot)) == docroot then
                  local loc = string.sub(p, string.len(docroot))
                  fh:write("\tlocation ~ ^", loc, " {\n")
                  fh:write("\t\tauth_basic \"", escape(opts.pwprotect[pwpath].title), "\";\n")
                  fh:write("\t\tauth_basic_user_file ", opts.path, "/conf/.htpasswd;\n")
                  if opts.pwprotect[pwpath].users[1] ~= "*" then
                    -- limit access to specific user(s)
                    fh:write("\t\tif ($remote_user !~ ^(")
                    for i in pairs(opts.pwprotect[pwpath].users) do
                      if i > 1 then fh:write("|") end
                      fh:write(escapePCRE(opts.pwprotect[pwpath].users[i]))
                    end
                    fh:write(")?$) {\n")
                    fh:write("\t\t\treturn 403;\n")
                    fh:write("\t\t}\n")
                  end

                  -- special case for password-protected root directory
                  if loc == '/' or loc:sub(1, 13) == '/.well-known/' then
                    fh:write("\t\tlocation ~ ^/.well-known/(.*/)?([^/]*)$ {\n")
                    fh:write("\t\t\terror_page 470 = @wellknown;\n")
                    fh:write("\t\t\tif (-f ", LC.web.getWebRoot(), "/.well-known/htdocs/$1$host.$2) { return 470; }\n")
                    fh:write("\t\t}\n")
                  end

                  fh:write("\t\tproxy_pass\thttp://127.0.0.1;\n")
                  local i, s
                  for i,s in pairs(PROXY_PARAMS) do
                    if type(s) == 'table' then
                      local j, t
                      for j,t in pairs(s) do fh:write("\t\t", i, "\t", t, ";\n") end
                    else
                      fh:write("\t\t", i, "\t", s, ";\n")
                    end
                  end

                  fh:write("\t}\n")
                end
              end
              fh:write("\n")
            end


            fh:write("}\n\n")
          end
        end

        break -- while(true)
      end
    end -- for...
  end -- if type(opts.vhosts) == "table"

  -- always write PHP-FCGI base config options, eg. to kill processes when PHP gets disabled
  local user = opts.name:lower()
  if opts.user then user = opts.user end
  fh:write("# PHP-FCGI configuration - DO NOT MODIFY!\n")
  fh:write("# NGINX_FCGI_USER=", user, "\n")
  fh:write("# NGINX_FCGI_CONF=", opts.path, "/conf/sockets\n")

  if vhostCount == 0 then
    fh:write("# No IPs/Domains configured for this vHost.\n")
  else
    if has_php then
      local idx = ""
      for i in pairs(phpVersionsFCGI) do
        local suffix = "." .. i
        if i == "" or i == LC.web.getPHPDefaultCode() then suffix = "" end

        fh:write("# NGINX_FCGI", idx, "_SOCKET=", opts.path, "/conf/sockets/nginx-php-fcgi", suffix, ".sock\n")
        fh:write("# NGINX_FCGI", idx, "_CHILDREN=5\n")
        fh:write("# NGINX_FCGI", idx, "_MAX_REQUESTS=1000\n")
        if LC.fs.is_file(opts.path .. "/.php5/php.ini") then
          fh:write("# NGINX_FCGI", idx, "_INI_PATH=", opts.path, "/.php5\n")
        else
          fh:write("# NGINX_FCGI", idx, "_INI_PATH=", opts.path, "/conf/")
          if i == "" then
            fh:write(LC.web.getPHPDefaultCode())
          else
            fh:write(i)
          end
          fh:write("\n")
        end

        local p
        for p in pairs(LCS.web.phpVersions) do
          if LCS.web.phpVersions[p].code == i or (i == "" and LCS.web.phpVersions[p].code == LC.web.getPHPDefaultCode()) then
            fh:write("# NGINX_FCGI", idx, "_BIN=", LCS.web.phpVersions[p].bin, "\n")
            break
          end
        end

        fh:write("\n")

        if not LC.fs.is_dir(opts.path .. "/conf/sockets") then
          LC.fs.mkdir(opts.path .. "/conf/sockets")
          LC.fs.setperm(opts.path .. "/conf/sockets", 755, opts.user, opts.group)
        end

        if idx == "" then
          idx = 1
        else
          idx = idx + 1
        end
      end -- for i, key in ipairs(phpVersions)
    end -- if opts.hasPHP and opts.hasPHP > 0

  end

  -- close (temporary) config file
  LC.liveconfig.writeFooter(fh)
  fh:close()

  -- rename config file (atomic)
  LC.fs.rename(tempfile, filename)

  if vhostCount == 0 then
    -- disable configuration, remove config file
    if cfg["disable_cmd"] ~= nil then
      LC.exec(cfg["disable_cmd"] .. " " .. opts["name"] .. ".conf")
    elseif cfg["enabled_dir"] ~= nil and vhostpath ~= cfg["enabled_dir"] and LC.fs.is_file(cfg["enabled_dir"] .. "/" .. opts["name"] .. ".conf") then
      -- remove link manually:
      os.remove(cfg["enabled_dir"] .. "/" .. opts["name"] .. ".conf")
    end
    os.remove(filename)

    -- stop PHP-FCGI for this customer
    if LC.fs.is_file("/etc/init.d/nginx-php-fcgi") then
      LC.exec("/etc/init.d/nginx-php-fcgi stop " .. opts.name)
    end
  else
    -- if not done yet, enable config
    if cfg["enable_cmd"] ~= nil then
      LC.exec(cfg["enable_cmd"] .. " " .. opts["name"] .. ".conf")
    elseif cfg["enabled_dir"] ~= nil and vhostpath ~= cfg["enabled_dir"] and not LC.fs.is_file(cfg["enabled_dir"] .. "/" .. opts["name"] .. ".conf") then
      -- link manually:
      LC.exec("ln -s " .. vhostpath .. "/" .. opts["name"] .. ".conf " .. cfg["enabled_dir"] .. "/" .. opts["name"] .. ".conf")
    end

    -- (re)start or stop PHP-FCGI for this customer
    if LC.fs.is_file("/etc/init.d/nginx-php-fcgi") then
      if has_php then
        LC.exec("/etc/init.d/nginx-php-fcgi restart " .. opts.name)
      else
        LC.exec("/etc/init.d/nginx-php-fcgi stop " .. opts.name)
      end
    end
  end

  -- reload web server configuration
  LC.timeout.set('nginx.reload', 10, 60)

  return true

end

-- ---------------------------------------------------------------------------
-- reload()
--
-- Reload NGINX configuration
-- ---------------------------------------------------------------------------
function reload()
  local retry
  local cmd

  LC.log.print(LC.log.DEBUG, "nginx.reload() called")
  -- get configuration
  local cfg = LC.web.getConfig('nginx')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "nginx.reload(): no configuration for 'nginx' available!?")
    return
  end

  if flag_restart then
    -- get restart command
    if cfg.restart_cmd == nil then
      LC.log.print(LC.log.ERR, "nginx.reload(): no restart command for 'nginx' available!?")
      return
    end
    cmd = cfg.restart_cmd
  else
    -- get reload command
    if cfg.reload_cmd == nil then
      LC.log.print(LC.log.ERR, "nginx.reload(): no reload command for 'nginx' available!?")
      return
    end
    cmd = cfg.reload_cmd
  end

  for retry = 0, 4 do
    -- reload/restart service
    local ret, out, err = LC.exec(cmd)

    if ret == 0 then
      flag_restart = false
      break
    end

    -- check for some common errors
    if string.match(err, "systemctl status nginx") then ret, err = LC.exec("systemctl status -l nginx.service") end
    local n = string.match(err, "you should increase server_names_hash_bucket_size: (%d+)")
    if n then
      -- increase server_names_hash_bucket_size
      local hashconf = cfg.configpath .. "/conf.d/hash.conf"
      local fh, msg = io.open(hashconf , "w")
      if fh == nil then
        LC.log.print(LC.log.ERR, "Can't open '", hashconf, "' for writing: ", msg)
        return false, "Can't open '" .. hashconf .. "' for writing: " .. msg
      end
      -- double size of server_names_hash_bucket_size
      local newsize = tonumber(n) * 2
      fh:write("server_names_hash_bucket_size ", newsize, ";\n")
      fh:close()
      LC.log.print(LC.log.INFO, "nginx.reload(): trying again with increased server_names_hash_bucket_size")
    else
      -- unknown error / don't know how to handle...
      break
    end
  end -- for retry...

end

-- ---------------------------------------------------------------------------
-- deleteAccount()
--
-- Delete account/vHost
-- ---------------------------------------------------------------------------
function deleteAccount(cfg, name)
  local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
  local filename = vhostpath .. "/" .. name .. ".conf"

  if not LC.fs.is_file(filename) then return end

  -- disable configuration, remove config file
  if cfg["disable_cmd"] ~= nil then
    LC.exec(cfg["disable_cmd"] .. " " .. name .. ".conf")
  elseif cfg["enabled_dir"] ~= nil and vhostpath ~= cfg["enabled_dir"] and LC.fs.is_file(cfg["enabled_dir"] .. "/" .. name .. ".conf") then
    -- remove link manually:
    os.remove(cfg["enabled_dir"] .. "/" .. name .. ".conf")
  end
  os.remove(filename)

  -- stop PHP-FCGI for this customer
  if LC.fs.is_file("/etc/init.d/nginx-php-fcgi") then
    LC.exec("/etc/init.d/nginx-php-fcgi stop " .. name)
  end

  -- reload web server configuration
  LC.timeout.set('nginx.reload', 10, 60)

end

-- ---------------------------------------------------------------------------
-- updateResolver()
--
-- Update file /etc/nginx/conf.d/resolver.conf according to "nameserver"
-- settings in /etc/resolv.conf
-- ---------------------------------------------------------------------------
function updateResolver()
  local fh, status, msg, line

  -- get configuration
  local cfg = LC.web.getConfig('nginx')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "nginx.reload(): no configuration for 'nginx' available!?")
    return
  end

  local v_major, v_minor, v_patch = string.match(cfg.binversion, "^(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)
  local ipv6_support = true
  if v_major < 1 or (v_major == 1 and (v_minor < 2 or (v_minor == 2 and v_patch < 2) or (v_minor == 3 and v_patch < 1))) then
    ipv6_support = false
  end

  -- get list of nameservers from /etc/resolv.conf:
  fh, msg = io.open("/etc/resolv.conf", "r")
  if fh == nil then
    LC.log.print(LC.log.ERR, "Can't open '/etc/resolv.conf' for reading: ", msg)
    return false, "Can't open '/etc/resolv.conf' for reading: " .. msg
  end

  local resolvers = {}

  while true do
    line = fh:read()
    if line == nil then break end
    local ns = line:match("^%s*nameserver%s+([0-9a-f.:]+).*$")
    if ns ~= nil then
      if ns:find(":") ~= nil then
        -- IPv6 address
        if ipv6_support then
          resolvers[#resolvers+1] = "[" .. ns .. "]"
        end
      else
        resolvers[#resolvers+1] = ns
      end
    end
  end

  fh:close()

  -- if no nameservers are defined at /etc/resolv.conf, user localhost (as by specification)
  if #resolvers == 0 then resolvers[1] = '127.0.0.1' end

  -- create new resolver configuration file
  local configpath = cfg["configpath"]
  local configfile = configpath .. "/conf.d/resolver.conf"
  fh, msg = io.open(configfile .. ".tmp", "w")
  if fh == nil then
    LC.log.print(LC.log.ERR, "Can't open '", configfile, ".tmp' for writing: ", msg)
    return false, "Can't open '" .. configfile .. ".tmp' for writing: " .. msg
  end

  -- write header
  LC.liveconfig.writeHeader(fh)

  fh:write("\n# list of DNS resolvers, created from /etc/resolv.conf\n")
  fh:write("# (NGINX can't use system resolver at runtime)\n")

  fh:write("resolver");
  for i in pairs(resolvers) do
    fh:write(" ", resolvers[i])
  end
  fh:write(";\n\n")

  -- write footer
  LC.liveconfig.writeFooter(fh)

  fh:close()

  -- rename temporary file
  LC.fs.rename(configfile .. ".tmp", configfile)

  return true

end

-- <EOF>----------------------------------------------------------------------
